# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2024- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr
#
#
# pyvkfft script with info about VkFFT version and CUDA/OpenCL support


import numpy as np
from pyvkfft.version import git_version, vkfft_git_version, __version__

vkfft_ver = None
try:
    from pyvkfft.cuda import has_pycuda, has_cupy, vkfft_version as cu_vkfft_version, \
        cuda_compile_version, cuda_runtime_version, cuda_driver_version

    vkfft_ver = cu_vkfft_version()
except (ImportError, OSError):
    has_cupy, has_pycuda = False, False

try:
    import pyopencl as cl

    has_opencl = True
    from pyvkfft.opencl import vkfft_version as cl_vkfft_version

    if vkfft_ver is None:
        vkfft_ver = cl_vkfft_version()

except (ImportError, OSError):
    has_opencl = False


def main():
    print(f"pyvkfft version: {__version__:12s} [git: {git_version()}]")
    print(f"VkFFT version:   {vkfft_ver:12s} [git: {vkfft_git_version()}]")
    print(f"numpy version:   {np.__version__:12s}")

    print(f"\nCUDA support: {has_pycuda or has_cupy}")
    if has_cupy or has_pycuda:
        print(f"  CUDA driver version:  {cuda_driver_version()}")
        print(f"  CUDA runtime version: {cuda_runtime_version()}")
        print(f"  CUDA compile version: {cuda_compile_version()}")

        if has_pycuda:
            from pycuda import VERSION_TEXT
            print(f"  pycuda available: {has_pycuda} , version={VERSION_TEXT}")
        else:
            print(f"  pycuda available: {has_pycuda}")
        if has_cupy:
            from cupy._version import __version__ as cp_ver

            print(f"  cupy available:   {has_cupy} , version={cp_ver}")
        else:
            print(f"  cupy available:   {has_cupy}")
        # Devices
        if has_pycuda:
            import pycuda.autoprimaryctx
            import pycuda.driver as cu_drv
            print(f"  #CUDA devices:   {cu_drv.Device.count()} (pycuda)")
            for i in range(cu_drv.Device.count()):
                print(f"       {i}: {cu_drv.Device(i).name()}")
        elif has_cupy:
            import cupy as cp
            print(f"  #CUDA devices:   {cp.cuda.runtime.getDeviceCount()} (cupy)")
            for i in range(cp.cuda.runtime.getDeviceCount()):
                print(f"       {i}: {cp.cuda.runtime.getDeviceProperties(i)['name'].decode()}")

    print(f"\nOpenCL support: {has_opencl}")
    if has_opencl:
        from pyopencl.version import VERSION_TEXT
        print(f"PyOpenCL version: {VERSION_TEXT}")
        print(f"  OpenCL platform and devices (GPU only):")
        for p in cl.get_platforms():
            nb_gpu = sum([d.type & cl.device_type.GPU for d in p.get_devices()])
            print(f"    platform: {p.name}")
            print(f"      Vendor:  {p.get_info(cl.platform_info.VENDOR)}")
            print(f"      Version: {p.get_info(cl.platform_info.VERSION)}")
            print(f"      #GPU devices: {nb_gpu}:")
            for d in p.get_devices():
                if d.type & cl.device_type.GPU:
                    print(f"        {d.name}")
                    try:
                        # CL_DEVICE_BOARD_NAME_AMD                        0x4038
                        print(f"          Board Name (AMD): {d.get_info(16440)}")
                    except cl.LogicError:
                        pass
                    print(f"          Version:          {d.get_info(cl.device_info.VERSION)}")
                    print(f"          Driver version:   {d.get_info(cl.device_info.DRIVER_VERSION)}")
                    print(f"          float64 support:  {'cl_khr_fp64' in d.get_info(cl.device_info.EXTENSIONS)}")
                    print(f"          float16 support:  {'cl_khr_fp16' in d.get_info(cl.device_info.EXTENSIONS)}")



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2022- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr
#
#
# script to run a long multi-test accuracy suite

import os
import glob
import platform
import argparse
import sys
import itertools


def make_parser():
    epilog = "Examples:\n" \
             "   pyvkfft-test-suite --gpumem 32 --backend pycuda --gpu V100\n\n" \
             "This will run the full test suite with a maximum of 20 parallel\n" \
             "process (based on the available memory), using pycuda on a V100 GPU\n\n" \
             "   pyvkfft-test-suite --gpumem 32 --backend pycuda --gpu V100 --transform c2c --ndim 1 2\n\n" \
             "This will run the test suite only for 1D and 2D C2C transforms"
    parser = argparse.ArgumentParser(prog='pyvkfft-test-suite', epilog=epilog,
                                     description='Run a complete test suite including the basic test, '
                                                 'followed by systematic tests for C2C, R2C, DCT, 1D, 2D and 3D\n'
                                                 'single and double precision, in and out-of-place, '
                                                 'different normalisations, with or without LUT (single),\n'
                                                 'radix and non-radix, plus a few asymmetric 2D and 3D tests.\n\n'
                                                 'Note that the full test suite typically takes 24 to 30 hours.\n\n'
                                                 'To merge the html files at the end, use:\n'
                                                 '      cat pyvkfft-test1*.html > pyvkfft-test.html ',
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('--gpu', action='store',
                        help="Name (or sub-string) of the GPU to use. If not given, "
                             "the first available will be used")
    parser.add_argument('--nproc', action='store',
                        help="Maximum number of parallel process to use to speed up tests. "
                             "This number will be decreased for larger arrays (e.g. 3D), but "
                             "it should be checked to avoid memory errors. A good value"
                             "for 32 GB is 20 processes.",
                        default=10, type=int)
    parser.add_argument('--gpumem', action='store',
                        help="Available GPU memory.",
                        default=16, type=int)
    parser.add_argument('--backend', action='store', required=True,
                        help="GPU backend",
                        choices=['pycuda', 'cupy', 'pyopencl'])
    parser.add_argument('--opencl_platform', action='store', default=None,
                        help="Name (or sub-string) of the opencl platform to use (case-insensitive)")
    parser.add_argument('--transform', action='store', nargs='+',
                        help="Transforms to test (defaults to all)",
                        default=['c2c', 'r2c', 'dct', 'dst'], choices=['c2c', 'r2c', 'dct', 'dst'])
    parser.add_argument('--radix', action='store_true',
                        help="Use this option to only test radix transforms")
    parser.add_argument('--single', action='store_true',
                        help="Use this option to only test single precision transforms")
    parser.add_argument('--ndim', action='store', nargs='+',
                        help="Number of dimensions for the tests. Several values can be given, "
                             "e.g. 1 2 3",
                        default=[1, 2, 3], type=int, choices=[1, 2, 3])
    parser.add_argument('--fast-random', action='store', default=None, type=int,
                        help="Use this option to run a random percentage of the systematic tests, "
                             "for faster results. A number (percentage) between 5 and 100 is required.")
    parser.add_argument('--dry-run', action='store_true',
                        help="Perform a dry-run, printing the commands for each sub-test.")
    parser.add_argument('--skip', action='store_true',
                        help="Use this option to skip tests which have already been performed (based"
                             "on parsing the html files in the current directory).")
    return parser


def main():
    parser = make_parser()
    args = parser.parse_args()

    backend = args.backend
    dry_run = args.dry_run
    gpu = args.gpu
    gpumem = args.gpumem
    nproc0 = args.nproc
    opencl_platform = args.opencl_platform
    radix = args.radix
    vdim = args.ndim

    if args.fast_random is not None:
        assert 5 <= args.fast_random <= 100, "--fast-random must be between 5 and 100"

    # Used for skipping already performed tests
    nhtml = 0
    ct_test = 0
    if args.skip:
        nhtml = len(glob.glob('pyvkfft-test1[0-8]*.html')) - 1

    # Basic test
    com = "pyvkfft-test --nproc %d --html --range-mb 0 4100 --backend %s --gpu %s" % (nproc0, backend, gpu)
    if opencl_platform is not None:
        com += ' --opencl_platform ' + opencl_platform
    if dry_run:
        if not (args.skip and nhtml >= 1):
            print(com)
        else:
            print(f'Skipping already performed test: {com}')
    else:
        if args.skip and nhtml >= 1:
            print(f'Skipping already performed test: {com}')
        else:
            if os.system(com) == 2:
                # Keyboard interrupt (why 2 ?)
                print("Aborting test suite")
                sys.exit(1)
    ct_test += 1

    # systematic tests
    vtransform = []
    if 'c2c' in args.transform:
        vtransform.append('        ')
    if 'r2c' in args.transform:
        vtransform.append(' --r2c  ')
    if 'dct' in args.transform:
        for i in range(1, 4 + 1):
            vtransform.append(' --dct %d' % i)
    if 'dst' in args.transform:
        for i in range(1, 4 + 1):
            vtransform.append(' --dst %d' % i)
    vnorm = [' --norm 1', ' --norm 0']
    vlut = ['', ' --lut']
    if args.single:
        vprec = ['']
    else:
        vprec = ['', ' --double']
    vradix = [' --radix'] if radix else [' --radix', ' --bluestein']
    vinplace = ['', ' --inplace']
    for radix, norm, lut, inplace, prec in itertools.product(vradix, vnorm, vlut, vinplace, vprec):
        if ' --lut' in lut and 'double' in prec:
            continue
        for transform in vtransform:
            if ('dct' in transform or 'dst' in transform) and '0' in norm:
                continue
            for ndim in vdim:
                n1 = 3 if ('dct 4' in transform or 'dst 4' in transform) else 2
                if ndim == 1:
                    n2 = 100000 if 'radix' in radix else 10000
                elif ndim == 2:
                    n2 = 4500
                else:  # ndim==3
                    n2 = 550
                # Estimate maximum memory usage
                mem = n2 ** ndim * 8
                # Assume 48 or 64 KB of shared memory to see if Vkfft needs a temp buffer.
                # For single precision and 64 KB, x-axis can be up to 8192 single upload,
                # 4096 for the other axes.
                if 'opencl' in backend:
                    b = (n2 * (16 if 'double' in prec else 8) / (48 * 1024)) * (1 + int(ndim > 1)) >= 1
                else:
                    b = (n2 * (16 if 'double' in prec else 8) / (64 * 1024)) * (1 + int(ndim > 1)) >= 1

                if 'double' in prec:
                    mem *= 2 + int(b)
                elif b:
                    mem *= 2

                if 'inplace' not in inplace:
                    mem *= 2
                if 'r2c' in transform:
                    # DCT/DST are solved as C2C systems, so if a buffer is allocated
                    # it will be 2x larger than the original data
                    mem /= 2
                elif 'dct 1' in transform or 'dst 1' in transform:
                    # System is computed as C2C of size 2N-2
                    mem *= 2
                if 'bluestein' in radix:
                    # Bluestein can require 2x memory
                    # (this over-estimates memory for Rader..)
                    mem *= 2

                mem += 200 * 1024 ** 2  # context memory
                # *2 is a margin for errors, e.g. to account for temp buffer
                nproc1 = gpumem / (mem / 1024 ** 3 * 2)
                if nproc1 < 1:
                    n2 *= nproc1 ** (1 / ndim)
                    nproc1 = 1
                else:
                    nproc1 = int(nproc1)
                nproc = max(1, min(nproc1, nproc0))
                com = 'pyvkfft-test --systematic --backend %s' % backend
                if gpu is not None:
                    com += ' --gpu %s' % gpu
                com += ' --graph --html --max-nb-tests 0'
                com += ' --nproc %2d --ndim %d --range %d %6d' % (nproc, ndim, n1, n2)
                com += transform + radix + inplace + prec + lut + norm + ' --range-mb 0 4100'
                if opencl_platform is not None:
                    com += ' --opencl_platform ' + opencl_platform
                if args.fast_random is not None:
                    com += ' --fast-random %d' % args.fast_random
                if dry_run:
                    if args.skip and nhtml > ct_test:
                        print(f'Skipping already performed test: {com}')
                    else:
                        print(com)
                        # os.system(com + ' --dry-run')
                else:
                    if args.skip and nhtml > ct_test:
                        print(f'Skipping already performed test: {com}')
                    else:
                        if os.system(com) == 2:
                            # Keyboard interrupt (why 2 ?)
                            print("Aborting test suite")
                            sys.exit(1)
                ct_test += 1

    # Last, run a few 2D and 3D tests where the lengths can differ,
    # and radix and Bluestein transforms are mixed.
    v = []
    if 2 in vdim:
        v += [(' --radix', '', '', '', 2, 2, 4500, ' --range-nd-narrow 0.04 8 --radix-max-pow 3'),
              (' --radix', ' --lut', '', '', 2, 2, 4500, ' --range-nd-narrow 0.04 8 --radix-max-pow 3'),
              (' --radix', '', ' --inplace', '', 2, 2, 4500, ' --range-nd-narrow 0.04 8 --radix-max-pow 3')]
        if not args.single:
            v += [(' --radix', '', '', ' --double', 2, 2, 4500, ' --range-nd-narrow 0.04 8 --radix-max-pow 3')]
        if not radix:
            v += [('', '', '', '', 2, 1008, 1040, ' --range-nd-narrow 0.04 8'),
                  ('', '', '', '', 2, 2032, 2064, ' --range-nd-narrow 0.04 8'),
                  ('', '', '', '', 2, 4080, 4112, ' --range-nd-narrow 0.04 8'),
                  ]
    if 3 in vdim:
        v += [(' --radix', '', '', '', 3, 2, 150, ' --range-nd-narrow 0.04 8 --radix-max-pow 3'),
              (' --radix', ' --lut', '', '', 3, 2, 150, ' --range-nd-narrow 0.04 8 --radix-max-pow 3'),
              (' --radix', '', ' --inplace', '', 3, 2, 150, ' --range-nd-narrow 0.04 8 --radix-max-pow 3')]
        if not args.single:
            v += [(' --radix', '', '', ' --double', 3, 2, 150, ' --range-nd-narrow 0.04 8 --radix-max-pow 3')]
        if not radix:
            v += [('', '', '', '', 3, 120, 140, ' --range-nd-narrow 0.04 8'),
                  ]

    for transform in vtransform:
        norm = ' --norm 1'
        for radix, lut, inplace, prec, ndim, n1, n2, rn in v:
            mem = n2 ** ndim * 8
            if 'double' in prec:
                mem *= 2
            if 'inplace' not in inplace:
                mem *= 2
            if 'dct' in transform or 'r2c' in transform or 'dst' in transform:
                mem /= 2
            nproc1 = gpumem // (mem / 1024 ** 3 * 1.5)
            nproc = max(1, min(nproc1, nproc0))
            com = 'pyvkfft-test --systematic --backend %s' % backend
            if gpu is not None:
                com += ' --gpu %s' % gpu
            com += ' --graph --html --max-nb-tests 0'
            com += ' --nproc %2d --ndim %d --range %d %6d' % (nproc, ndim, n1, n2)
            com += transform + radix + inplace + prec + lut + norm + rn + ' --range-mb 0 4100'
            if opencl_platform is not None:
                com += ' --opencl_platform ' + opencl_platform
            if args.fast_random is not None:
                com += ' --fast-random %d' % args.fast_random

            if dry_run:
                if args.skip and nhtml > ct_test:
                    print(f'Skipping already performed test: {com}')
                else:
                    print(com)
                    # os.system(com + ' --dry-run')
            else:
                if args.skip and nhtml > ct_test:
                    print(f'Skipping already performed test: {com}')
                else:
                    if os.system(com) == 2:
                        # Keyboard interrupt (why 2 ?)
                        print("Aborting test suite")
                        sys.exit(1)
            ct_test += 1


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2021- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr

import os
import platform
import sysconfig
import ctypes
import warnings
from enum import Enum
import numpy as np
from . import config

# np.complex32 does not exist yet https://github.com/numpy/numpy/issues/14753
complex32 = np.dtype([('re', np.float16), ('im', np.float16)])

# Type to pass array size, omit and batch as int arrays
ctype_int_size_p = np.ctypeslib.ndpointer(dtype='long', ndim=1, flags='C_CONTIGUOUS')


class VkFFTResult(Enum):
    """ VkFFT error codes from vkFFT.h """
    VKFFT_SUCCESS = 0
    VKFFT_ERROR_MALLOC_FAILED = 1
    VKFFT_ERROR_INSUFFICIENT_CODE_BUFFER = 2
    VKFFT_ERROR_INSUFFICIENT_TEMP_BUFFER = 3
    VKFFT_ERROR_PLAN_NOT_INITIALIZED = 4
    VKFFT_ERROR_NULL_TEMP_PASSED = 5
    VKFFT_ERROR_INVALID_PHYSICAL_DEVICE = 1001
    VKFFT_ERROR_INVALID_DEVICE = 1002
    VKFFT_ERROR_INVALID_QUEUE = 1003
    VKFFT_ERROR_INVALID_COMMAND_POOL = 1004
    VKFFT_ERROR_INVALID_FENCE = 1005
    VKFFT_ERROR_ONLY_FORWARD_FFT_INITIALIZED = 1006
    VKFFT_ERROR_ONLY_INVERSE_FFT_INITIALIZED = 1007
    VKFFT_ERROR_INVALID_CONTEXT = 1008
    VKFFT_ERROR_INVALID_PLATFORM = 1009
    VKFFT_ERROR_ENABLED_saveApplicationToString = 1010,
    VKFFT_ERROR_EMPTY_FFTdim = 2001
    VKFFT_ERROR_EMPTY_size = 2002
    VKFFT_ERROR_EMPTY_bufferSize = 2003
    VKFFT_ERROR_EMPTY_buffer = 2004
    VKFFT_ERROR_EMPTY_tempBufferSize = 2005
    VKFFT_ERROR_EMPTY_tempBuffer = 2006
    VKFFT_ERROR_EMPTY_inputBufferSize = 2007
    VKFFT_ERROR_EMPTY_inputBuffer = 2008
    VKFFT_ERROR_EMPTY_outputBufferSize = 2009
    VKFFT_ERROR_EMPTY_outputBuffer = 2010
    VKFFT_ERROR_EMPTY_kernelSize = 2011
    VKFFT_ERROR_EMPTY_kernel = 2012
    VKFFT_ERROR_EMPTY_applicationString = 2013,
    VKFFT_ERROR_UNSUPPORTED_RADIX = 3001
    VKFFT_ERROR_UNSUPPORTED_FFT_LENGTH = 3002
    VKFFT_ERROR_UNSUPPORTED_FFT_LENGTH_R2C = 3003
    VKFFT_ERROR_UNSUPPORTED_FFT_LENGTH_R2R = 3004
    VKFFT_ERROR_UNSUPPORTED_FFT_OMIT = 3005
    VKFFT_ERROR_FAILED_TO_ALLOCATE = 4001
    VKFFT_ERROR_FAILED_TO_MAP_MEMORY = 4002
    VKFFT_ERROR_FAILED_TO_ALLOCATE_COMMAND_BUFFERS = 4003
    VKFFT_ERROR_FAILED_TO_BEGIN_COMMAND_BUFFER = 4004
    VKFFT_ERROR_FAILED_TO_END_COMMAND_BUFFER = 4005
    VKFFT_ERROR_FAILED_TO_SUBMIT_QUEUE = 4006
    VKFFT_ERROR_FAILED_TO_WAIT_FOR_FENCES = 4007
    VKFFT_ERROR_FAILED_TO_RESET_FENCES = 4008
    VKFFT_ERROR_FAILED_TO_CREATE_DESCRIPTOR_POOL = 4009
    VKFFT_ERROR_FAILED_TO_CREATE_DESCRIPTOR_SET_LAYOUT = 4010
    VKFFT_ERROR_FAILED_TO_ALLOCATE_DESCRIPTOR_SETS = 4011
    VKFFT_ERROR_FAILED_TO_CREATE_PIPELINE_LAYOUT = 4012
    VKFFT_ERROR_FAILED_SHADER_PREPROCESS = 4013
    VKFFT_ERROR_FAILED_SHADER_PARSE = 4014
    VKFFT_ERROR_FAILED_SHADER_LINK = 4015
    VKFFT_ERROR_FAILED_SPIRV_GENERATE = 4016
    VKFFT_ERROR_FAILED_TO_CREATE_SHADER_MODULE = 4017
    VKFFT_ERROR_FAILED_TO_CREATE_INSTANCE = 4018
    VKFFT_ERROR_FAILED_TO_SETUP_DEBUG_MESSENGER = 4019
    VKFFT_ERROR_FAILED_TO_FIND_PHYSICAL_DEVICE = 4020
    VKFFT_ERROR_FAILED_TO_CREATE_DEVICE = 4021
    VKFFT_ERROR_FAILED_TO_CREATE_FENCE = 4022
    VKFFT_ERROR_FAILED_TO_CREATE_COMMAND_POOL = 4023
    VKFFT_ERROR_FAILED_TO_CREATE_BUFFER = 4024
    VKFFT_ERROR_FAILED_TO_ALLOCATE_MEMORY = 4025
    VKFFT_ERROR_FAILED_TO_BIND_BUFFER_MEMORY = 4026
    VKFFT_ERROR_FAILED_TO_FIND_MEMORY = 4027
    VKFFT_ERROR_FAILED_TO_SYNCHRONIZE = 4028
    VKFFT_ERROR_FAILED_TO_COPY = 4029
    VKFFT_ERROR_FAILED_TO_CREATE_PROGRAM = 4030
    VKFFT_ERROR_FAILED_TO_COMPILE_PROGRAM = 4031
    VKFFT_ERROR_FAILED_TO_GET_CODE_SIZE = 4032
    VKFFT_ERROR_FAILED_TO_GET_CODE = 4033
    VKFFT_ERROR_FAILED_TO_DESTROY_PROGRAM = 4034
    VKFFT_ERROR_FAILED_TO_LOAD_MODULE = 4035
    VKFFT_ERROR_FAILED_TO_GET_FUNCTION = 4036
    VKFFT_ERROR_FAILED_TO_SET_DYNAMIC_SHARED_MEMORY = 4037
    VKFFT_ERROR_FAILED_TO_MODULE_GET_GLOBAL = 4038
    VKFFT_ERROR_FAILED_TO_LAUNCH_KERNEL = 4039
    VKFFT_ERROR_FAILED_TO_EVENT_RECORD = 4040
    VKFFT_ERROR_FAILED_TO_ADD_NAME_EXPRESSION = 4041
    VKFFT_ERROR_FAILED_TO_INITIALIZE = 4042
    VKFFT_ERROR_FAILED_TO_SET_DEVICE_ID = 4043
    VKFFT_ERROR_FAILED_TO_GET_DEVICE = 4044
    VKFFT_ERROR_FAILED_TO_CREATE_CONTEXT = 4045
    VKFFT_ERROR_FAILED_TO_CREATE_PIPELINE = 4046
    VKFFT_ERROR_FAILED_TO_SET_KERNEL_ARG = 4047
    VKFFT_ERROR_FAILED_TO_CREATE_COMMAND_QUEUE = 4048
    VKFFT_ERROR_FAILED_TO_RELEASE_COMMAND_QUEUE = 4049
    VKFFT_ERROR_FAILED_TO_ENUMERATE_DEVICES = 4050
    VKFFT_ERROR_FAILED_TO_GET_ATTRIBUTE = 4051
    VKFFT_ERROR_FAILED_TO_CREATE_EVENT = 4052


def _library_path(basename):
    if platform.system() == 'Windows':
        # We patched build_ext so the module is a .so and not a dll
        ext = '.so'
    else:
        ext = sysconfig.get_config_var('EXT_SUFFIX')
    return os.path.join(os.path.dirname(__file__) or os.path.curdir, basename + ext)


def load_library(basename):
    return ctypes.cdll.LoadLibrary(_library_path(basename))


def primes(n):
    """ Returns the prime decomposition of n as a list.
    This only remains as a useful function, but VkFFT
    allows any prime decomposition, even if performance
    can be better for prime numbers <=13.
    """
    v = [1]
    assert (n > 0)
    i = 2
    while i * i <= n:
        while n % i == 0:
            v.append(i)
            n //= i
        i += 1
    if n > 1:
        v.append(n)
    return v


def radix_gen(nmax, radix, even=False, exclude_one=True, inverted=False,
              nmin=None, max_pow=None, r2r=False):
    """
    Generate an array of integers which are only multiple of powers
    of base integers, e.g. 2**N1 * 3**N2 * 5**N3 etc...

    :param nmax: the maximum integer to return (included)
    :param radix: the list/tuple of base integers - which don't need
        to be primes
    :param even: if True, only return even numbers
    :param exclude_one: if True (the default), exclude 1
    :param inverted: if True, the returned array will only include
        integers which are NOT in the form 2**N1 * 3**N2 * 5**N3...
    :param nmin: if not None, the integer values returned will be >=nmin
    :param max_pow: if not None, the N1, N2, N3... powers (for sizes
        in the form 2**N1 * 3**N2 * 5**N3) will be at max equal to this
        value, which allows to reduce the number of generated sizes
        while testing all base radixes
    :param r2r: if r2r=4, assume we try to generate radix sizes for
        a DCT1 or DST1 transform, which are computed as a C2C
        transform of size 2N-2.
        If r2r=4, assume we try to generate radix sizes for a
        DCT4 or DST4, which are performed as a C2C transform
        of size n/2 (if n is even) or n (if odd).
    :return: the numpy array of integers, sorted
    """
    a = np.ones(1, dtype=np.int64)
    for i in range(len(radix)):
        if max_pow is None:
            tmp = np.arange(int(np.floor(np.log(nmax) / np.log(radix[i]))) + 1)
        else:
            tmp = np.arange(min(max_pow, int(np.floor(np.log(nmax) / np.log(radix[i])))) + 1)
        a = a * radix[i] ** tmp[:, np.newaxis]
        a = a.flatten()
        a = a[a <= nmax]
    if r2r == 1:
        # We need 2N-2 to be radix, so use generated n+1
        a = a[a > 1] + 1
    elif r2r == 4:
        a[a % 2 == 0] //= 2
        a = np.unique(a)
    a = a[a <= nmax]

    if inverted:
        b = np.arange(nmax + 1)
        b[a] = 0
        a = b.take(np.nonzero(b)[0])
    if even:
        a = a[(a % 2) == 0]
    if nmin is not None:
        a = a[a >= nmin]
    a.sort()
    a = a[a > 0]  # weird cases with r2r=1 or 4
    if len(a):
        if exclude_one and a[0] == 1:
            return a[1:]
    return a


def radix_gen_n(nmax, max_size, radix, ndim=None, even=False, exclude_one=True, inverted=False,
                nmin=None, max_pow=None, range_nd_narrow=None, min_size=0, r2r=False):
    """
    Generate a list of array shape with integers which are only multiple
    of powers of base integers, e.g. 2**N1 * 3**N2 * 5**N3 etc...,
    for each of the dimensions, and with a maximum size.
    Note that this can generate a large number of sizes.

    :param nmax: the maximum value for the length of each dimension (included)
    :param max_size: the maximum size (number of elements) for the array.
    :param radix: the list/tuple of base integers - which don't need
        to be primes. If None, all sizes are allowed
    :param ndim: the number of dimensions allowed. If None, 1D, 2D and 3D
        shapes are mixed.
    :param even: if True, only return even numbers
    :param exclude_one: if True (the default), exclude 1
    :param inverted: if True, the returned array will only include
        integers which are NOT in the form 2**N1 * 3**N2 * 5**N3...
    :param nmin: if not None, the integer values returned will be >=nmin
    :param max_pow: if not None, the N1, N2, N3... powers (for sizes
        in the form 2**N1 * 3**N2 * 5**N3) will be at max equal to this
        value, which allows to reduce the number of generated sizes
        while testing all base radixes
    :param range_nd_narrow: if a tuple of values (drel, dabs) is given,
        with drel within [0;1], for dimensions>1,
        in an array of shape (s0, s1, s2), the difference of lengths
        with respect to the first dimension cannot be larger than
        min(drel * s0, dabs). This allows to reduce the number
        of shapes tested. With drel=dabs=0, all dimensions must
        have identical lengths.
    :param min_size: the minimum size (number of elements). This can be
        used to separate large array tests and use a larger number of
        parallel process for smaller ones.
    :param r2r: if r2r=4, assume we try to generate radix sizes for
        a DCT1 or DST1 transform, which are computed as a C2C
        transform of size 2N-2.
        If r2r=4, assume we try to generate radix sizes for a
        DCT4 or DST4, which are performed as a C2C transform
        of size n/2 (if n is even) or n (if odd).
    :return: the list of array shapes.
    """
    v = []
    if radix is None:
        if even:
            if nmin is None:
                n0 = np.arange(2, min(nmax, max_size), 2)
            else:
                n0 = np.arange(nmin + nmin % 2, min(nmax, max_size), 2)
        else:
            if nmin is None:
                n0 = np.arange(2, min(nmax, max_size))
            else:
                n0 = np.arange(nmin, min(nmax, max_size))
    else:
        n0 = radix_gen(nmax, radix, even=even, exclude_one=exclude_one, inverted=inverted,
                       nmin=nmin, max_pow=max_pow, r2r=r2r)

    if ndim is None or ndim in [1, 12, 123]:
        idx = np.nonzero((n0 <= max_size) * (n0 >= min_size))[0]
        if len(idx):
            v += list(zip(n0.take(idx)))
    if ndim is None or ndim in [2, 12, 123]:
        vidx = list(range(0, len(n0), 1000))
        if vidx[-1] != len(n0):
            vidx.append(len(n0))
        for i1 in range(len(vidx) - 1):
            l01 = n0[vidx[i1]:vidx[i1 + 1]]
            for i2 in range(len(vidx) - 1):
                l2 = n0[vidx[i2]:vidx[i2 + 1]][:, np.newaxis]
                s = (l01 * l2).flatten()
                l1, l2 = (l01 + np.zeros_like(l2)).flatten(), (l2 + np.zeros_like(l01)).flatten()
                tmp = (s <= max_size) * (s >= min_size)
                if range_nd_narrow is not None:
                    drel, dabs = range_nd_narrow
                    m = np.maximum(dabs, l1 * drel)
                    tmp = np.logical_and(tmp, abs(l1 - l2) <= m)
                idx = np.nonzero(tmp)[0]
                if len(idx):
                    v += list(zip(l1.take(idx), l2.take(idx)))
    if ndim is None or ndim in [3, 123]:
        vidx = list(range(0, len(n0), 100))
        if vidx[-1] != len(n0):
            vidx.append(len(n0))
        for i1 in range(len(vidx) - 1):
            l01 = n0[vidx[i1]:vidx[i1 + 1]]
            for i2 in range(len(vidx) - 1):
                l02 = n0[vidx[i2]:vidx[i2 + 1], np.newaxis]
                for i3 in range(len(vidx) - 1):
                    l3 = n0[vidx[i3]:vidx[i3 + 1], np.newaxis, np.newaxis]
                    # print(i1, i2, i3, l1.shape, l2.shape, l3.shape)
                    s = (l01 * l02 * l3).flatten()
                    l1, l2, l3 = (l01 + np.zeros_like(l02) + np.zeros_like(l3)).flatten(), \
                        (l02 + np.zeros_like(l01) + np.zeros_like(l3)).flatten(), \
                        (l3 + np.zeros_like(l01) + np.zeros_like(l02)).flatten()
                    tmp = (s <= max_size) * (s >= min_size)
                    if range_nd_narrow is not None:
                        drel, dabs = range_nd_narrow
                        m = np.maximum(dabs, l1 * drel)
                        tmp = np.logical_and(tmp, (abs(l1 - l2) <= m)
                                             * (abs(l1 - l3) <= m))
                    idx = np.nonzero(tmp)[0]
                    if len(idx):
                        v += list(zip(l1.take(idx), l2.take(idx), l3.take(idx)))
    return v


def strides_nonzero(strides):
    """
    Fix the strides for an array, if one is zero it should be set to the
    smallest stride between the next and previous nonzero stride.
    """
    if strides is None or np.isscalar(strides):
        return strides
    n = len(strides)
    strides = list(strides)
    for i in range(n):
        s = strides[i]
        if s == 0:
            # use the maximum previous or next nonzero stride
            m1, m2 = 0, 0
            for ii in range(i + 1, n - 1):
                if strides[ii] > 0:
                    m1 = strides[ii]
                    break
            for ii in range(i - 1, -1, -1):
                if strides[ii] > 0:
                    m2 = strides[ii]
                    break
            strides[i] = min(m1, m2)
    return strides


def calc_transform_axes(shape, axes=None, ndim=None, strides=None):
    """ Compute the final shape of the array to be passed
    to VkFFT, and the axes for which the transform should
    be skipped.
    By collapsing non-transformed consecutive axes and using batch transforms,
    it is possible to support larger dimensions (the limit is set at
    compilation time).

    :param shape: the initial shape of the data array. Note that this shape
        should be in the usual numpy order, i.e. the fastest axis is
        listed last. e.g. (nz, ny, nx)
    :param axes: the axes to be transformed. if None, all axes
        are transformed, or up to ndim.
    :param ndim: the number of dimensions for the transform. If None,
        the number of axes is used
    :param strides: the array strides. If None, a C-order is assumed
        with the fastest axes along the last dimensions (numpy default)
    :return: (shape, n_batch, skip_axis, ndim, axes0) with the shape after collapsing
        consecutive non-transformed axes (padded with ones if necessary,
        with the order adequate for VkFFT i.e. (nx, ny, nz,...),
        the batch size (e.g. 5 if shape=(5,16,16) and ndim=2),
        the list of booleans indicating which axes should
        be skipped, and the number of transform axes. Finally, axes0
        is returned as a list of transformed axes, before any axis collapsing.
    """
    # reverse order to have list as (nx, ny, nz,...)
    shape1 = list(reversed(list(shape)))
    n = len(shape)
    if np.isscalar(axes):
        axes = [axes]
    if ndim is None:
        if axes is None:
            ndim1 = len(shape)
            axes = list(range(ndim1))
    else:
        ndim1 = ndim
        if axes is not None:
            if ndim1 != len(axes):
                raise RuntimeError("The number of transform axes does not match ndim:", axes, ndim)

    # Axes with VkFFT order
    if axes is not None:
        axes1 = [n - 1 - ax if ax >= 0 else -ax - 1 for ax in axes]
    else:
        # will be modified to take fastest axes if F-ordered
        axes1 = list(range(ndim1))

    if strides is not None and n > 1:
        # The fast axis must come first for VkFFT.
        # Careful with the special case of axes with size 1 (stride does not increase or is 0...)
        # Try with shape=(3,1), F and C-ordered as a corner case...
        s0 = strides_nonzero(strides)
        reorder_shape, reorder_axes = False, False
        if not np.all([s0[i] >= s0[i + 1] for i in range(n - 1)]):
            # Array is F-ordered, need to reorder shape & axes
            if axes is None:
                # Using ndim, so axes1 is already OK
                reorder_shape = True
            else:
                reorder_shape, reorder_axes = True, True
        elif axes is not None and (np.sum([s > 1 for s in shape]) == 1):
            # Case from hell: e.g. shape=(3,1). Strides won't help, so use axes
            if axes1[-1] == n - 1:  # same as axes[-1]==0
                reorder_shape, reorder_axes = True, True
        if reorder_shape:
            shape1 = shape1[::-1]
        if reorder_axes:
            axes1 = [n - 1 - ax if ax >= 0 else -ax - 1 for ax in axes1]

    # List of non-transformed axes
    skip_axis = [True] * len(shape1)
    for i in axes1:
        skip_axis[i] = False

    if np.all([shape1[ax] == 1 for ax in axes1]):
        raise RuntimeError(f"No axis is actually transformed: shape={shape} ndim={ndim} "
                           f"axes={axes} strides={strides} vkfft_shape={shape1} "
                           f"vkfft_axes={axes1} vkfft_skip={skip_axis}")

    # Collapse non-transform axes when possible
    i = 0
    while i <= len(shape1) - 2:
        if skip_axis[i] and skip_axis[i + 1]:
            shape1[i] *= shape1[i + 1]
            shape1.pop(i + 1)
            skip_axis.pop(i + 1)
        else:
            i += 1

    # Fix ndim so skipped axes are counted
    ndim1 = len(shape1) - list(reversed(skip_axis)).index(False)

    # For VkFFT > cc2b427, all dimensions beyond the
    # transformed axes should be in n_batch
    n_batch = 1
    for i in range(ndim1, len(shape1)):
        n_batch *= shape1[i]
        shape1[i] = 1
        # Axes beyond ndim are marked skipped
        skip_axis[i] = True

    if axes is None:
        # Return the actual axes transformed
        axes = [n - 1 - ax if ax >= 0 else -ax - 1 for ax in axes1]

    return shape1, n_batch, skip_axis, ndim1, axes


def check_vkfft_result(res, shape=None, dtype=None, ndim=None, inplace=None,
                       norm=None, r2c=None, dct=None, dst=None, axes=None, backend=None,
                       strides=None, vkfft_shape=None, vkfft_skip=None, vkfft_nbatch=None):
    """
    Check VkFFTResult code.

    :param res: the result code from launching a transform.
    :param shape: shape of the array
    :param dtype: data type of the array
    :param ndim: number of transform dimensions
    :param inplace: True or False
    :param norm: 0 or1 or "ortho"
    :param r2c: True or False
    :param dct: False, 1, 2, 3 or 4
    :param dst: False, 1, 2, 3 or 4
    :param axes: transform axes
    :param backend: the backend
    :param strides: the array strides
    :param vkfft_shape: the shape passed to VkFFT
    :param vkfft_skip: the skipped axis list passed to VkFFT
    :param vkfft_nbatch: VkFFT batch parameter
    :raises RuntimeError: if res != 0
    """
    if isinstance(res, ctypes.c_int):
        res = res.value
    if res != 0:
        s = ""
        if r2c:
            s += "R2C "
        elif dct:
            s += "DCT%d " % dct
        elif dst:
            s += "DST%d " % dst
        else:
            s += "C2C "
        if r2c and inplace and shape is not None:
            tmp = list(shape)
            tmp[-1] -= 2
            shstr = str(tuple(tmp)).replace(" ", "")
            if ",)" in shstr:
                s += shstr.replace(",)", "+2)") + " "
            else:
                s += shstr.replace(")", "+2)") + " "
        else:
            s += str(shape).replace(" ", "") + " "
        if dtype is not None:
            s += str(dtype) + " "
        if axes is not None:
            s += str(axes).replace(" ", "") + " "
        if strides is not None:
            s += "strides=" + str(strides).replace(" ", "") + " "
        if ndim is not None:
            s += "%dD " % ndim
        if inplace:
            s += "inplace "
        if norm:
            s += "norm=%s " % str(norm)
        if vkfft_shape is not None or vkfft_skip is not None or vkfft_nbatch is not None:
            s += "[VkFFT:"
            if vkfft_shape is not None:
                s += " shape= " + str(vkfft_shape).replace(" ", "")
            if vkfft_skip is not None:
                s += " skip=" + str([int(sk) for sk in vkfft_skip]).replace(" ", "")
            if vkfft_nbatch is not None:
                s += f" nbatch={vkfft_nbatch}"
            s += "] "
        if backend is not None:
            s += "[%s]" % backend
        try:
            r = VkFFTResult(res)
            raise RuntimeError("VkFFT error %d: %s %s" % (res, r.name, s))
        except ValueError:
            raise RuntimeError("VkFFT error %d (unknown) %s" % (res, s))


class VkFFTApp:
    """
    VkFFT application interface implementing a FFT plan, base implementation
    handling functions and paremeters common to the CUDA and OpenCL backends.
    """

    def __init__(self, shape, dtype: type, ndim=None, inplace=True, norm=1,
                 r2c=False, dct=False, dst=False, axes=None, strides=None,
                 r2c_odd=False, **kwargs):
        """
        Init function for the VkFFT application.

        :param shape: the shape of the array to be transformed. The number
            of dimensions of the array can be larger than the FFT dimensions.
        :param dtype: the numpy dtype of the source array (can be complex64 or complex128)
        :param ndim: the number of dimensions to use for the FFT. By default,
            uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
            array to perform a batched 3D FFT on all the layers. The FFT
            is always performed along the last axes if the array's number
            of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
            on the x and y axes for ndim=2, etc.. Unless axes are given.
        :param inplace: if True (the default), performs an inplace transform and
            the destination array should not be given in fft() and ifft().
        :param norm: if 0 (unnormalised), every transform multiplies the L2
            norm of the array by its size (or the size of the transformed
            array if ndim<d.ndim).
            if 1 (the default) or "backward", the inverse transform divides
            the L2 norm by the array size, so FFT+iFFT will keep the array norm.
            if "ortho", each transform will keep the L2 norm, but that will
            involve an extra read & write operation.
        :param r2c: if True, will perform a real->complex transform, where the
            complex destination is a half-hermitian array.
            For an inplace transform, if the transformed data shape is (...,nx),
            the input float array should have a shape of (..., nx+2), the last
            two columns being ignored in the input data, and the resulting
            complex array (using pycuda's GPUArray.view(dtype=np.complex64) to
            reinterpret the type) will have a shape (..., nx//2 + 1).
            For an out-of-place transform, if the input (real) shape is (..., nx),
            the output (complex) shape should be (..., nx//2+1).
            Note 1: the above shape changes are true for C-contiguous arrays;
            generally the axis which is halved by the R2C transform always is
            the fast axis -with a stride of 1 element. For F-contiguous arrays
            this will be the first dimension instead of the last.
            Note 2:for C2R transforms with ndim>=2, the source (complex) array
            is modified.
        :param dct: used to perform a Direct Cosine Transform (DCT) aka a R2R transform.
            An integer can be given to specify the type of DCT (1, 2, 3 or 4).
            if dct=True, the DCT type 2 will be performed, following scipy's convention.
        :param dst: used to perform a Direct Sine Transform (DST) aka a R2R transform.
            An integer can be given to specify the type of DST (1, 2, 3 or 4).
            if dst=True, the DST type 2 will be performed, following scipy's convention.
        :param axes: a list or tuple of axes along which the transform should be made.
            if None, the transform is done along the ndim fastest axes, or all
            axes if ndim is None. For R2C transforms, the fast axis must be
            transformed.
        :param strides: the array strides - needed if not C-ordered.
        :param r2c_odd: this should be set to True to perform an inplace r2c/c2r
            transform with an odd-sized fast (x) axis.
            Explanation: to perform a 1D inplace transform of an array with 100
            elements, the input array should have a 100+2 size, resulting in
            a half-Hermitian array of size 51. If the input data has a size
            of 101, the input array should also be padded to 102 (101+1), and
            the resulting half-Hermitian array also has a size of 51. A
            flag is thus needed to differentiate the cases of 100+2 or 101+1.

        :raises RuntimeError:  if the transform dimensions or data type
            are not allowed by VkFFT.
        """
        self.app = None
        self.config = None
        if ((dct or dst) and r2c) or (dct and dst):
            raise RuntimeError("R2C, DCT and DST are mutually exclusive")
        if (r2c or dct or dst) and dtype not in [np.float16, np.float32, np.float64]:
            raise RuntimeError("R2C, DST or DCT selected but input type is not real")
        self.fast_axis = len(shape) - 1  # default for C-ordered if strides is None
        if strides is not None:
            # Getting the real fast axis can be tricky. Lots of corner cases,
            # as the stride can be zero for axes of size 1...
            if axes is not None and np.sum([sh > 1 for sh in shape]) == 1:
                # Strides won't help, so use the last transformed axis listed
                # as fast axis-following numpy convention
                self.fast_axis = axes[-1]
            else:
                s0 = strides_nonzero(strides)
                if not np.all([s0[i] >= s0[i + 1] for i in range(len(shape) - 1)]):
                    # F-ordered array
                    self.fast_axis = 0
        if r2c and axes is not None:
            if self.fast_axis not in axes and -len(shape) + self.fast_axis not in axes:
                raise RuntimeError(f"the fast axis must be transformed for R2C"
                                   f" [axes={axes}, strides={strides}], "
                                   f"fast axis={self.fast_axis}/{-len(shape) + self.fast_axis}]")
        if r2c and inplace:
            if shape[self.fast_axis] % 2:
                raise RuntimeError(f"For an inplace R2C/C2R transform, the supplied array shape {shape} "
                                   f"must be even along the fast (x) axis. If the transform size nx is "
                                   f"even, two buffer elements should be added to the end of the axis."
                                   f"If it is odd, only one element should be added. In both cases, "
                                   f"the complex half-Hermitian array size will be nx//2+1.")
        # Get the final shape passed to VkFFT, collapsing non-transform axes
        # as necessary. The calculated shape has 4 dimensions (nx, ny, nz, n_batch).
        self.shape, self.n_batch, self.skip_axis, self.ndim, self.axes0 = \
            calc_transform_axes(shape, axes, ndim, strides)
        # original shape (without collapsed non-transformed axes)
        self.shape0 = shape
        self.strides0 = strides
        self.inplace = inplace
        self.r2c = r2c
        self.r2c_odd = r2c_odd
        if dct is False:
            self.dct = 0
        elif dct is True:
            self.dct = 2
        else:
            self.dct = dct
        if dst is False:
            self.dst = 0
        elif dst is True:
            self.dst = 2
        else:
            self.dst = dst
        if dct and self.dct < 1 or self.dct > 4:
            raise RuntimeError("Only DCT of types 1, 2, 3 and 4 are allowed")
        if dst and self.dst < 1 or self.dst > 4:
            raise RuntimeError("Only DST of types 1, 2, 3 and 4 are allowed")

        # These parameters will be filled in by the different backends
        # Size of the temp buffer allocated by VkFFT
        self.tmp_buffer_nbytes = 0
        # 0 or 1 for each axis, only if the Bluestein algorithm is used (same length as self.shape)
        self.use_bluestein_fft = None
        # number of axis upload per dimension (same length as self.shape)
        self.nb_axis_upload = None

        # Experimental parameters. Not much difference is seen, so don't document this,
        # VkFFT default parameters seem fine.

        # force callback version of R2C and R2R (DCT/DST) algorithms for all usecases (0 - off, 1 - on)
        # this is normally activated automatically by VkFFT for odd sizes.
        if "forceCallbackVersionRealTransforms" in kwargs:
            self.forceCallbackVersionRealTransforms = kwargs["forceCallbackVersionRealTransforms"]
        else:
            self.forceCallbackVersionRealTransforms = -1

        # disables unshuffling of Four step algorithm. Requires tempbuffer allocation (0 - off, 1 - on)
        if "disableReorderFourStep" in kwargs:
            self.disableReorderFourStep = kwargs["disableReorderFourStep"]
        else:
            self.disableReorderFourStep = -1

        # uint64_t coalescedMemory - number of bytes to coalesce per one transaction.
        # For Nvidia and AMD is equal to 32, Intel is equal to 64. Going to work regardless,
        # but if specified by the user correctly, the performance will be higher. Default 64
        # for other GPUs. For half-precision should be multiplied by two. Should be a power of two.
        if "coalescedMemory" in kwargs:
            self.coalescedMemory = kwargs["coalescedMemory"]
        else:
            self.coalescedMemory = -1

        # uint64_t numSharedBanks - configure the number of shared banks on the target GPU.
        # Default 32. Minor performance boost as it solves shared memory conflicts for
        # the power of two systems.
        if "numSharedBanks" in kwargs:
            self.numSharedBanks = kwargs["numSharedBanks"]
        else:
            self.numSharedBanks = -1

        # uint64_t aimThreads - try to aim all kernels at this amount of threads.
        # Gains/losses are not predictable, just a parameter to play with
        # (it is not guaranteed that the target kernel will use that many threads).
        # Default 128.
        if "aimThreads" in kwargs:
            self.aimThreads = kwargs["aimThreads"]
        else:
            self.aimThreads = -1

        # uint64_t performBandwidthBoost - try to reduce coalesced number by a factor of X
        # to get bigger sequence in one upload for strided axes.
        # Default: -1(inf) for DCT, 2 for Bluestein’s algorithm (or -1 if DCT), 0 otherwise
        if "performBandwidthBoost" in kwargs:
            self.performBandwidthBoost = kwargs["performBandwidthBoost"]
        else:
            self.performBandwidthBoost = -1

        # uint64_t registerBoost - specify if the register file size is bigger than
        # shared memory and can be used to extend it X times (on Nvidia 256KB register
        # file can be used instead of 32KB of shared memory, set this constant to 4 to
        # emulate 128KB of shared memory). Default 1 - no over-utilization. In Vulkan,
        # OpenCL and Level Zero it is set to 4 on Nvidia GPUs, to 2 if the driver
        # shows 64KB or more of shared memory on AMD, to 2 if the driver shows less
        # than 64KB of shared memory on AMD, to 1 if the driver shows 64KB or more
        # of shared memory on Intel, to 2 if the driver shows less than 64KB of shared
        # memory on Intel.
        if "registerBoost" in kwargs:
            self.registerBoost = kwargs["registerBoost"]
        else:
            self.registerBoost = -1

        # uint64_t registerBoostNonPow2 - specify if register over-utilization should
        # be used on non-power of 2 sequences. Default 0, set to 1 to enable.
        if "registerBoostNonPow2" in kwargs:
            self.registerBoostNonPow2 = kwargs["registerBoostNonPow2"]
        else:
            self.registerBoostNonPow2 = -1

        # uint64_t registerBoost4Step - specify if register file over-utilization
        # should be used in big sequences (>2^14), same definition as registerBoost.
        # Default 1.
        if "registerBoost4Step" in kwargs:
            self.registerBoost4Step = kwargs["registerBoost4Step"]
        else:
            self.registerBoost4Step = -1

        # # uint64_t maxComputeWorkGroupCount[3] - how many workgroups can be launched
        # # at one dispatch. Automatically derived from the driver, can be artificially
        # # lowered. Then VkFFT will perform a logical split and extension of the
        # # number of workgroups to cover the required range.
        # if "maxComputeWorkGroupCount" in kwargs:
        #     self.maxComputeWorkGroupCount = kwargs["maxComputeWorkGroupCount"]
        # else:
        #     self.maxComputeWorkGroupCount = (-1, -1, -1)
        #
        # # uint64_t maxComputeWorkGroupSize[3] - max dimensions of the workgroup.
        # # Automatically derived from the driver. Can be modified if there are
        # # some issues with the driver (as there were with ROCm 4.0, when it returned
        # # 1024 for maxComputeWorkGroupSize and actually supported only up to 256 threads).
        # if "maxComputeWorkGroupSize" in kwargs:
        #     self.maxComputeWorkGroupSize = kwargs["maxComputeWorkGroupSize"]
        # else:
        #     self.maxComputeWorkGroupSize = (-1, -1, -1)
        #
        # # uint64_t maxThreadsNum - max number of threads per block. Similar to maxCompute
        # # - WorkGroupSize, but aggregated. Automatically derived from the driver.
        # if "maxThreadsNum" in kwargs:
        #     self.maxThreadsNum = kwargs["maxThreadsNum"]
        # else:
        #     self.maxThreadsNum = -1
        #
        # # uint64_t sharedMemorySizeStatic - available for static allocation shared
        # # memory size, in bytes. Automatically derived from the driver.
        # if "sharedMemorySizeStatic" in kwargs:
        #     self.sharedMemorySizeStatic = kwargs["sharedMemorySizeStatic"]
        # else:
        #     self.sharedMemorySizeStatic = -1
        #
        # # uint64_t sharedMemorySize - available for allocation shared memory size,
        # # in bytes. VkFFT uses dynamic shared memory in CUDA/HIP as it allows for
        # # bigger allocations.
        # if "sharedMemorySize" in kwargs:
        #     self.sharedMemorySize = kwargs["sharedMemorySize"]
        # else:
        #     self.sharedMemorySize = -1
        #
        # # uint64_t sharedMemorySizePow2 - the power of 2 which is less or equal to
        # # sharedMemorySize, in bytes.
        # if "sharedMemorySizePow2" in kwargs:
        #     self.sharedMemorySizePow2 = kwargs["sharedMemorySizePow2"]
        # else:
        #     self.sharedMemorySizePow2 = -1

        # uint64_t warpSize - number of threads per warp/wavefront. Automatically derived
        # from the driver, but can be modified (can increase performance, though
        # unpredictable as defaults have good values). Must be a power of two.
        if "warpSize" in kwargs:
            self.warpSize = kwargs["warpSize"]
        else:
            self.warpSize = -1

        self.groupedBatch = [-1, -1, -1]
        if "groupedBatch" in kwargs:
            self.groupedBatch = list(kwargs["groupedBatch"])
            # In case less than 3 parameters where given
            self.groupedBatch[:len(kwargs["groupedBatch"])] = kwargs["groupedBatch"]

        if "useLUT" in kwargs:
            # useLUT=1 may be beneficial on platforms which have a low accuracy for
            # the native sincos functions.
            if kwargs["useLUT"] is None:
                self.use_lut = -1
            else:
                self.use_lut = kwargs["useLUT"]
        elif config.USE_LUT is not None:
            self.use_lut = config.USE_LUT
        else:
            self.use_lut = -1

        if "keepShaderCode" in kwargs:
            # This will print the compiled code if equal to 1
            self.keepShaderCode = kwargs["keepShaderCode"]
        else:
            self.keepShaderCode = -1

        if norm == "backward":
            norm = 1
        self.norm = norm

        # Precision: number of bytes per float
        if dtype in [np.float16, complex32]:
            self.precision = 2
        elif dtype in [np.float32, np.complex64]:
            self.precision = 4
        elif dtype in [np.float64, np.complex128]:
            self.precision = 8

    def __str__(self):
        """
        Get a string describing the VkFFTApp properties, e.g.:
          VkFFTApp[OpenCL]:(212,212+2)     R2C/s/i [RR] [11] buf=    0
        This includes VkFFTApp with the backend (OpenCL or CUDA) used, followed
        by the shape of the array (in python order),
        then the type of transform (C2C or R2C or DCT/DST), then
        h/s/d for half/single/double precisions,
        i or o for in or out-of-place transforms,
        [???] with a letter indicating for each axis if it uses a
        [r]adix, [B]luestein or [R]ader transform.
        Then [nnn] indicates how manu uploads are used per axis,
        and finally the size of the temporary buffer allocated by VkFFT
        is indicated, if any.
        """
        bufs = self.get_tmp_buffer_str()
        ft_type = self.get_algo_str()

        sh = self.shape.copy()
        sh[-1] *= self.n_batch
        naxup = ''.join([str(i) for i in self.get_nb_upload()])
        s = "VkFFTApp[OpenCL]:" if 'opencl' in str(self.__class__) else "VkFFTApp[CUDA]:  "
        s += f"{self.get_shape_str():15s}"
        if self.dct:
            s += f"DCT{self.dct}"
        elif self.dst:
            s += f"DST{self.dst}"
        elif self.r2c:
            s += " R2C"
        else:
            s += " C2C"
        s += {2: "/h", 4: "/s", 8: "/d"}[self.precision]
        s += '/i' if self.inplace else '/o'
        s += f" [{ft_type}] [{naxup}]"
        s += f" buf={bufs}"
        return s

    def _get_fft_scale(self, norm):
        """Return the scale factor by which an array must be multiplied to keep its L2 norm
        after a forward FT
        :param norm: the norm option for which the scale is computed, either 0 or 1
        :return: the scale factor, as a numpy float with the precision used for the fft
        """
        dtype = np.float32
        if self.precision == 8:
            dtype = np.float64
        elif self.precision == 2:
            dtype = np.float16
        s = 1
        ndim_real = 0
        for i in range(self.ndim):
            if not self.skip_axis[i]:
                s *= self.shape[i]
                ndim_real += 1
        s = np.sqrt(s)
        if self.r2c and self.inplace:
            # Note: this is still correct for non-C-ordered arrays since
            # self.shape has been re-ordered
            if self.r2c_odd:
                s *= np.sqrt((self.shape[0] - 1) / self.shape[0])
            else:
                s *= np.sqrt((self.shape[0] - 2) / self.shape[0])
        if self.dct or self.dst:
            s *= 2 ** (0.5 * ndim_real)
            if max(self.dct, self.dst) != 4:  # if dct is used, dst=0 and inversely
                warnings.warn("A DST or DCT type 2 or 3 cannot be strictly normalised, using approximation,"
                              " see https://en.wikipedia.org/wiki/Discrete_cosine_transform#DCT-II")
        if norm == 0 or norm == 1:
            return dtype(1 / s)
        elif norm == "ortho":
            return dtype(1)
        raise RuntimeError("Unknown norm choice !")

    def get_fft_scale(self):
        """Return the scale factor by which an array must be multiplied to keep its L2 norm
        after a forward FT
        """
        return self._get_fft_scale(self.norm)

    def _get_ifft_scale(self, norm):
        """Return the scale factor by which an array must be multiplied to keep its L2 norm
        after a backward FT
        :param norm: the norm option for which the scale is computed, either 0 or 1
        :return: the scale factor, as a numpy float with the precision used for the fft
        """
        dtype = np.float32
        if self.precision == 8:
            dtype = np.float64
        elif self.precision == 2:
            dtype = np.float16
        s = 1
        s_dct = 1  # used also for dst
        for i in range(self.ndim):
            if not self.skip_axis[i]:
                s *= self.shape[i]
                if self.dct or self.dst:
                    s_dct *= np.sqrt(2)
        s = np.sqrt(s)
        if self.r2c and self.inplace:
            if self.r2c_odd:
                s *= np.sqrt((self.shape[0] - 1) / self.shape[0])
            else:
                s *= np.sqrt((self.shape[0] - 2) / self.shape[0])
        r2r_type = max(self.dct, self.dst)
        if (self.dct or self.dst) and r2r_type != 4:
            warnings.warn("A DST or DCT type 2 or 3 cannot be strictly normalised, using approximation,"
                          " see https://en.wikipedia.org/wiki/Discrete_cosine_transform#DCT-II")
        if norm == 0:
            return dtype(1 / (s * s_dct))
        elif norm == 1:
            # Not sure why the difference in scale factors
            if r2r_type == 2:
                s_dct = s_dct ** 1
            elif r2r_type == 3:
                s_dct = s_dct ** 2
            elif r2r_type == 4:
                s_dct = s_dct ** 3
            return dtype(s * s_dct)
        elif norm == "ortho":
            return dtype(1)
        raise RuntimeError("Unknown norm choice !")

    def get_ifft_scale(self):
        """Return the scale factor by which an array must be multiplied to keep its L2 norm
        after a backward FT
        """
        return self._get_ifft_scale(self.norm)

    def get_tmp_buffer_nbytes(self):
        """
        Return the size (in bytes) of the temporary buffer allocated
        by VkFFT for the transform, if any.
        """
        return self.tmp_buffer_nbytes

    def get_tmp_buffer_str(self):
        """
        Get a string with the size of the temporary buffer allocated
        by VkFFT, e.g. '0', '123kB', '1.2GB', etc. Uses 6 chars.
        """
        b = self.tmp_buffer_nbytes
        return "    0  " if b == 0 else f"{b / 1024 ** 3:5.1f}GB" if b >= 1000 * 1024 ** 2 \
            else f"{b / 1024 ** 2:5.1f}MB" if b >= 1000 * 1024 \
            else f"{b / 1024 :5.1f}kB" if b >= 1000 else f"{b:6d}B "

    def get_algo_str(self, vkfft_axes=False):
        """
        Return a string indicating the type of algorithm used for each axis,
        either [r]adix, [B]luestein or [R]ader, or '-' if the axis
        is skipped.
        """
        if vkfft_axes:
            tmp = ''
            for i in range(len(self.shape)):
                if self.skip_axis[i]:
                    tmp += '-'
                elif self.is_radix_transform(i, vkfft_axes=True):
                    tmp += 'r'
                elif self.is_rader_transform(i, vkfft_axes=True):
                    tmp += 'R'
                else:
                    tmp += 'B'
            return tmp
        else:
            tmp = ''
            for i in range(len(self.shape0)):
                if self.skip_axis[self._get_vkfft_axes(i)]:
                    tmp += '-'
                elif self.is_radix_transform(i):
                    tmp += 'r'
                elif self.is_rader_transform(i):
                    tmp += 'R'
                else:
                    tmp += 'B'
            return tmp

    def get_shape_str(self, vkfft_axes=False):
        """
        Get a string with the shape of the array, including the +1
        or +2 for inplace r2c transforms.

        :param vkfft_axes: True to use the index relative to VkFFT
            axes, False (the default) otherwise.
            See _get_vkfft_axes() for details.
        """
        # TODO: handle non C-contiguous array (strides) for inplace r2c
        if vkfft_axes:
            sh = list(self.shape).copy()
        else:
            sh = list(self.shape0).copy()

        if self.r2c and self.inplace:
            # Need to figure out the fast axis for the +1 or +2
            if vkfft_axes:
                fast_axis = 0
            else:
                # _get_vkfft_axes() takes into account strides
                fast_axis = np.argmin(self._get_vkfft_axes())

            if self.r2c_odd:
                r2c_inplace_pad = 1
            else:
                r2c_inplace_pad = 2

            sh[fast_axis] -= r2c_inplace_pad
            tmp = [str(sh[i]) + f'+{r2c_inplace_pad}' if i == fast_axis
                   else str(sh[i]) for i in range(len(sh))]
            return f"({','.join(tmp)})"
        return f"({','.join([str(n) for n in sh])})"

    def _get_vkfft_axes(self, i0=None):
        """
        Get the index of an axis in self.shape, self.skip_axis,
        self.use_bluestein_fft, etc..., from the original index.
        This is used because consecutive non-transformed axes are collapsed (merged).
        Additionally, VkFFT indexes the dimensions
        as [nx ny nz] rather than [nz ny nx], which is also taken care of.
        Finally, this takes into account strides if the array is not C-contiguous,
        which will change the order of the axes stored for VkFFT.
        Example: an array of shape (10,8,6,4) and axes=[-4,-1] will be internally
        seen as an array of shape (4,48,10).

        :param i0: the index in the original numpy array shape (can be >=0 or <0)
        :return: the index used for the same axis with the VkFFT order, i.e.
            reversed and after collapsing non-transformed axes. If i0=None,
            the index of all dimensions is returned as a list
        """
        n = len(self.shape0)
        i1 = len(self.shape) - 1
        idx = [i1]
        for i in range(n - 1):
            if i in self.axes0 or i - n in self.axes0 or i + 1 in self.axes0 or i + 1 - n in self.axes0:
                i1 -= 1
            idx.append(i1)
        if self.strides0 is not None:
            idx = np.take(idx, np.argsort(self.strides0)[::-1])
        if i0 is None:
            return idx
        return idx[i0]

    def is_bluestein_transform(self, axis=None, vkfft_axes=False):
        """
        Return True if the transform used along a given axis uses
        Bluestein's algorithm, or False

        :param axis: the index of one axis. If None, a list of values
            for all the axis is returned
        :param vkfft_axes: True to use the index relative to VkFFT
            axes, False (the default) otherwise.
            See _get_vkfft_axes() for details.
        """
        b = self.use_bluestein_fft
        if vkfft_axes:
            if axis is not None:
                return bool(b[axis])
            else:
                return b
        else:
            if axis is not None:
                return b[self._get_vkfft_axes(axis)]
            else:
                return [b[self._get_vkfft_axes(i)] for i in range(len(self.shape0))]

    def is_rader_transform(self, axis=None, vkfft_axes=False):
        """
        Return True if the transform used along a given axis uses
        Rader's algorithm, or False

        :param axis: the index of one axis. If None, a list of values
            for all the axis is returned
        :param vkfft_axes: True to use the index relative to VkFFT
            axes, False (the default) otherwise.
            See _get_vkfft_axes() for details.
        """
        b = self.use_bluestein_fft
        t = self.skip_axis
        sh = list(self.shape).copy()
        if self.r2c and self.inplace:
            # Fast axis is always first in the VkFFT order
            if self.r2c_odd:
                sh[0] -= 1
            else:
                sh[0] -= 2

        if self.dct == 1 or self.dst == 1:
            # Transforms are mapped to a C2C of size 2N-2
            sh = [max(2 * n - 2, 1) for n in sh]  # avoid 2*(n=1)-2=0
        elif self.dct == 4 or self.dst == 4:
            # even mapped to C2C of half-size, else C2C of same size
            sh = [n if n % 2 else n // 2 for n in sh]
        r = [max(primes(n)) <= 13 for n in sh]

        if vkfft_axes:
            if axis is not None:
                return not (b[axis] or t[axis] or r[axis])
            else:
                return [not (b[i] or t[i] or r[i]) for i in range(len(b))]
        else:
            if axis is not None:
                i = self._get_vkfft_axes(axis)
                return not (b[i] or t[i] or r[i])
            else:
                return [not (b[self._get_vkfft_axes(i)] or
                             t[self._get_vkfft_axes(i)] or
                             r[self._get_vkfft_axes(i)]) for i in range(len(self.shape0))]

    def is_radix_transform(self, axis=None, vkfft_axes=False):
        """
        Return True if the transform used along a given axis uses
        a radix algorithm, or False

        :param axis: the index of one axis. If None, a list of values
            for all axes is returned
        :param vkfft_axes: True to use the index relative to VkFFT
            axes, False (the default) otherwise.
            See _get_vkfft_axes() for details.
        """
        t = self.skip_axis
        sh = list(self.shape).copy()
        if self.r2c and self.inplace:
            # Fast axis is always first in the VkFFT order
            if self.r2c_odd:
                sh[0] -= 1
            else:
                sh[0] -= 2

        if self.dct == 1 or self.dst == 1:
            # Transforms are mapped to a C2C of size 2N-2
            sh = [max(2 * n - 2, 1) for n in sh]  # avoid 2*(n=1)-2=0
        elif self.dct == 4 or self.dst == 4:
            # even mapped to C2C of half-size, else C2C of same size
            sh = [n if n % 2 else n // 2 for n in sh]
        r = [max(primes(n)) > 13 for n in sh]

        if vkfft_axes:
            if axis is not None:
                return not (t[axis] or r[axis])
            else:
                return [not (t[i] or r[i]) for i in range(len(t))]
        else:
            if axis is not None:
                i = self._get_vkfft_axes(axis)
                return not (t[i] or r[i])
            else:
                return [not (t[self._get_vkfft_axes(i)] or r[self._get_vkfft_axes(i)])
                        for i in range(len(self.shape0))]

    def get_nb_upload(self, axis=None, vkfft_axes=False):
        """
        Number of uploads for the transform along given axes - ideally 1
        so that each transform corresponds to 1 read and 1 write of the array.

        :param axis: the index of one axis. If None, a list of values
            for all axes is returned
        :param collapsed_axes: True to use the index relative to collapsed
            axes, False (the default) otherwise.
            See _get_vkfft_axes() for details.
        """
        n = self.nb_axis_upload
        if vkfft_axes:
            if axis is not None:
                return n[::-1][axis]
            else:
                return n[::-1]
        else:
            if axis is not None:
                i = self._get_vkfft_axes(axis)
                return n[i]
            else:
                return [n[self._get_vkfft_axes(i)] for i in range(len(self.shape0))]



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2021- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr
#

"""
Benchmark functions. These are implemented using separate process,
one for each test - this involves a fair amount of overhead, but avoids
any resource conflict, or issue with GPU contexts, deletion of cufft
plans (https://github.com/lebedov/scikit-cuda/issues/308), etc..
"""

import warnings
import os
import timeit
from multiprocessing import Process, Queue
import platform
from itertools import permutations
from time import localtime, strftime
import numpy as np
import matplotlib.pyplot as plt

from pyvkfft.version import vkfft_version, vkfft_git_version
from pyvkfft.base import primes

try:
    from cpuinfo import get_cpu_info
except ImportError:
    get_cpu_info = None


def get_cpu_name():
    if get_cpu_info is None:
        return "CPU"
    d = get_cpu_info()
    return f"{d['brand_raw']}"


# test for GPU packages in parallel process (slower but cleaner)

def _test_pyvkfft_cuda(q):
    try:
        import pycuda.autoinit
        import pycuda.driver as cuda
        import pycuda.gpuarray as cua
        from pycuda import curandom
        import pyvkfft.cuda
        from pyvkfft.cuda import VkFFTApp as cuVkFFTApp, cuda_compile_version, \
            cuda_driver_version, cuda_runtime_version
        q.put((True, cuda_compile_version(), cuda_driver_version(), cuda_runtime_version()))
    except:
        q.put((False, None, None, None))


def test_pyvkfft_cuda():
    """
    Test if pyvkfft_cuda is available. The test is made in a separate process.
    Also return the
    """
    q = Queue()
    p = Process(target=_test_pyvkfft_cuda, args=(q,))
    p.start()
    has_pyvkfft_cuda, cu_version_compile, cu_version_driver, cu_version_runtime = q.get()
    p.join()
    return has_pyvkfft_cuda, cu_version_compile, cu_version_driver, cu_version_runtime


def _test_pyvkfft_opencl(q):
    try:
        import pyopencl as cl
        import pyopencl.array as cla
        from pyopencl import clrandom
        import pyvkfft.opencl
        from pyvkfft.opencl import VkFFTApp as clVkFFTApp
        q.put(True)
    except:
        q.put(False)


def test_pyvkfft_opencl():
    """
    Test if pyvkfft_opencl is available. The test is made in a separate process.
    """
    q = Queue()
    p = Process(target=_test_pyvkfft_opencl, args=(q,))
    p.start()
    has_pyvkfft_opencl = q.get()
    p.join()
    return has_pyvkfft_opencl


def _test_skcuda(q):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            import pycuda.autoinit
            import pycuda.driver as cu_drv
            import pycuda.gpuarray as cua
            from pycuda import curandom
            import skcuda.fft as cu_fft
            from skcuda.cufft import _cufft_version as v
            q.put((True, "%d.%d.%d" % (v // 1000, (v // 100) % 100, v % 100)))
        except:
            q.put((False, None))


def test_skcuda():
    q = Queue()
    p = Process(target=_test_skcuda, args=(q,))
    p.start()
    has_skcuda, cufft_version = q.get()
    p.join()
    return has_skcuda, cufft_version


def _test_cupy(q):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            import cupy as cp
            with cp.cuda.Device(0).use():
                vd = cp.cuda.runtime.driverGetVersion()
                vr = cp.cuda.runtime.runtimeGetVersion()
                q.put((True, "%d.%d.%d" % (vd // 1000, (vd // 100) % 100, vd % 100),
                       "%d.%d.%d" % (vr // 1000, (vr // 100) % 100, vr % 100)))
        except:
            q.put((False, None))


def test_cupy():
    q = Queue()
    p = Process(target=_test_cupy, args=(q,))
    p.start()
    has_cupy, cuda_driver_version, cuda_runtime_version = q.get()
    p.join()
    return has_cupy, cuda_driver_version, cuda_runtime_version


def _test_gpyfft(q):
    """
    Test if scikit-cuda is available. The test is made in a separate process.
    """
    try:
        import pyopencl as cl
        import pyopencl.array as cla
        from pyopencl import clrandom
        import gpyfft
        q.put(True)
    except:
        q.put(False)


def test_gpyfft():
    """
    Test if gpyfft is available. The test is made in a separate process.
    """
    q = Queue()
    p = Process(target=_test_gpyfft, args=(q,))
    p.start()
    has_gpyfft = q.get()
    p.join()
    return has_gpyfft


def _bench_pyvkfft_opencl(q, sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None,
                          opencl_platform=None, args=None, inplace=True, r2c=False, dct=False, dst=False):
    import pyopencl as cl
    import pyopencl.array as cla
    from pyopencl import clrandom
    import pyvkfft.opencl
    from pyvkfft.opencl import VkFFTApp as clVkFFTApp

    if r2c or dct or dst:
        dtype = np.float64 if precision == 'double' else np.float32
    else:
        dtype = np.complex128 if precision == 'double' else np.complex64

    gpu_name_real = gpu_name
    platform_name_real = opencl_platform
    if 'PYOPENCL_CTX' in os.environ:
        cl_ctx = cl.create_some_context()
    else:
        cl_ctx = None
        has_pocl = False
        for p in cl.get_platforms():
            if opencl_platform is not None:
                if opencl_platform.lower() not in p.name.lower():
                    continue
            elif "portable" in p.name.lower():
                # Try to skip PoCL unless it was requested
                has_pocl = True
                continue
            for d in p.get_devices():
                if gpu_name is not None:
                    if gpu_name.lower() not in d.name.lower():
                        continue
                if d.type & cl.device_type.GPU == 0:
                    continue
                gpu_name_real = d.name
                platform_name_real = p.name
                # print("Selected OpenCL device: ", d.name)
                cl_ctx = cl.Context(devices=(d,))
                break
            if cl_ctx is not None:
                break
        if cl_ctx is None and opencl_platform is None and has_pocl:
            # Try again without excluding PoCL
            for p in cl.get_platforms():
                for d in p.get_devices():
                    if gpu_name is not None:
                        if gpu_name.lower() not in d.name.lower():
                            continue
                    if d.type & cl.device_type.GPU == 0:
                        continue
                    gpu_name_real = d.name
                    platform_name_real = p.name
                    # print("Selected OpenCL device: ", d.name)
                    cl_ctx = cl.Context(devices=(d,))
                    break
                if cl_ctx is not None:
                    break
    cq = cl.CommandQueue(cl_ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)
    dt = 0

    r2c_odd = False
    if r2c and inplace:
        # Pad with 1 or 2 elements for inplace r2c
        sh = list(sh)
        if sh[-1] % 2:
            r2c_odd = True
            sh[-1] += 1
        else:
            sh[-1] += 2
        sh = tuple(sh)

    d = clrandom.rand(cq, shape=sh, dtype=np.float32).astype(dtype)
    if inplace:
        d1 = d
    else:
        if r2c:
            sh1 = [n for n in d.shape]
            sh1[-1] = sh1[-1] // 2 + 1
            d1 = cla.empty(cq, shape=tuple(sh1), dtype=np.complex128 if precision == 'double' else np.complex64)
        else:
            d1 = cla.empty_like(d)
    try:
        kwargs = {}
        if args is not None:
            for k, v in args.items():
                if k in ["disableReorderFourStep", "coalescedMemory", "numSharedBanks",
                         "aimThreads", "performBandwidthBoost", "registerBoost",
                         "registerBoostNonPow2", "registerBoost4Step", "warpSize", "useLUT",
                         "groupedBatch", "forceCallbackVersionRealTransforms", "tune_config"]:
                    kwargs[k] = v
        app = clVkFFTApp(d.shape, dtype=dtype, queue=cq, ndim=ndim, inplace=inplace,
                         r2c=r2c, dct=dct, dst=dst, r2c_odd=r2c_odd, **kwargs)
        algo_str = app.get_algo_str()
        nup_str = ''.join(str(nup) for nup in app.get_nb_upload())
        vkfft_str = f"algo={algo_str} buf={app.get_tmp_buffer_str()} up={nup_str}"
        if 'tune_config' in kwargs:
            for k, v in kwargs['tune_config'].items():
                if k == "backend":
                    continue
                vkfft_str += f" {k}={getattr(app, k)}"
        for i in range(nb_repeat):
            cq.finish()
            t0 = timeit.default_timer()
            # Apparently OpenCL events don't always work. Need kernel events ?
            # start = cl.enqueue_marker(cq)
            for ii in range(nb_loop):
                d1 = app.fft(d, d1)
                d = app.ifft(d1, d)
            # end = cl.enqueue_marker(cq)
            # end.wait()
            # dt1 = 1e-9 * (start.profile.END - end.profile.END)
            cq.finish()
            dt1 = (timeit.default_timer() - t0) / nb_loop
            if dt == 0:
                dt = dt1
            elif dt1 < dt:
                dt = dt1
        # print("%4d %4dx%4d 2D FFT+iFFT dt=%6.2f ms %7.2f Gbytes/s [pyvkfft.opencl]  [nb=%4d]" %
        #      (nz, n, n, dt / nb * 1000, gbps, nb))
        gbps = d.nbytes * ndim * 2 * 2 / dt / 1024 ** 3
    except Exception:
        import traceback
        print(traceback.format_exc())
        gbps = 0
    results = {'dt': dt, 'gbps': gbps, 'gpu_name_real': gpu_name_real,
               'platform_name_real': platform_name_real,
               'vkfft_str': vkfft_str, 'algo_str': algo_str, 'nup_str': nup_str}
    if q is None:
        return results
    else:
        q.put(results)


def bench_pyvkfft_opencl(sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None,
                         opencl_platform=None, args=None, serial=False,
                         inplace=True, r2c=False, dct=False, dst=False):
    if serial:
        return _bench_pyvkfft_opencl(None, sh, precision, ndim, nb_repeat, nb_loop, gpu_name,
                                     opencl_platform, args, inplace=inplace,
                                     r2c=r2c, dct=dct, dst=dst)
    q = Queue()
    p = Process(target=_bench_pyvkfft_opencl, args=(q, sh, precision, ndim, nb_repeat, nb_loop, gpu_name,
                                                    opencl_platform, args, inplace, r2c, dct, dst))
    p.start()
    try:
        results = q.get(timeout=20)
    except:
        results = {'dt': 0, 'gbps': 0, 'gpu_name_real': None, 'platform_name_real': None,
                   'vkfft_str': None}
    p.join()
    return results


def _bench_pyvkfft_cuda(q, sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None, args=None,
                        inplace=True, r2c=False, dct=False, dst=False):
    import pycuda.autoprimaryctx  # See https://github.com/lebedov/scikit-cuda/issues/330#issuecomment-1125471345
    import pycuda.driver as cu_drv
    import pycuda.gpuarray as cua
    from pycuda import curandom
    import pyvkfft.cuda
    from pyvkfft.cuda import VkFFTApp as cuVkFFTApp

    if r2c or dct or dst:
        dtype = np.float64 if precision == 'double' else np.float32
    else:
        dtype = np.complex128 if precision == 'double' else np.complex64

    if gpu_name is None:
        gpu_name_real = pycuda.autoprimaryctx.device.name()
        cu_ctx = pycuda.autoprimaryctx.context
    else:
        for i in range(cu_drv.Device.count()):
            d = cu_drv.Device(i)
            if gpu_name.lower() in d.name().lower():
                gpu_name_real = d.name()
                cu_ctx = d.retain_primary_context()
                break
    cu_ctx.push()
    dt = 0

    r2c_odd = False
    if r2c and inplace:
        # Pad with 1 or 2 elements for inplace r2c
        sh = list(sh)
        if sh[-1] % 2:
            r2c_odd = True
            sh[-1] += 1
        else:
            sh[-1] += 2
        sh = tuple(sh)

    d = curandom.rand(shape=sh, dtype=np.float32).astype(dtype)

    if inplace:
        d1 = d
    else:
        if r2c:
            sh1 = [n for n in d.shape]
            sh1[-1] = sh1[-1] // 2 + 1
            d1 = cua.empty(shape=tuple(sh1), dtype=np.complex128 if precision == 'double' else np.complex64)
        else:
            d1 = cua.empty_like(d)

    try:
        kwargs = {}
        if args is not None:
            for k, v in args.items():
                if k in ["disableReorderFourStep", "coalescedMemory", "numSharedBanks",
                         "aimThreads", "performBandwidthBoost", "registerBoost",
                         "registerBoostNonPow2", "registerBoost4Step", "warpSize", "useLUT",
                         "groupedBatch", "tune_config"]:
                    kwargs[k] = v
        app = cuVkFFTApp(d.shape, dtype=dtype, ndim=ndim, inplace=inplace,
                         r2c=r2c, dct=dct, dst=dst, r2c_odd=r2c_odd, **kwargs)
        algo_str = app.get_algo_str()
        nup_str = ''.join(str(nup) for nup in app.get_nb_upload())
        vkfft_str = f"algo={algo_str} buf={app.get_tmp_buffer_str()} up={nup_str}"
        if 'tune_config' in kwargs:
            for k, v in kwargs['tune_config'].items():
                if k == "backend":
                    continue
                vkfft_str += f" {k}={getattr(app, k)}"
        start = cu_drv.Event()
        stop = cu_drv.Event()
        for i in range(nb_repeat):
            cu_ctx.synchronize()
            start.record()
            for ii in range(nb_loop):
                d1 = app.fft(d, d1)
                d = app.ifft(d1, d)
            stop.record()
            cu_ctx.synchronize()
            dt1 = stop.time_since(start) / 1000 / nb_loop
            if dt == 0:
                dt = dt1
            elif dt1 < dt:
                dt = dt1
        # print("%4d %4dx%4d 2D FFT+iFFT dt=%6.2f ms %7.2f Gbytes/s [pyvkfft.cuda]    [nb=%4d]" %
        #      (nz, n, n, dt / nb * 1000, gbps, nb))
        del app
        gbps = d.nbytes * ndim * 2 * 2 / dt / 1024 ** 3
    except:
        gbps = 0
    cu_ctx.pop()
    results = {'dt': dt, 'gbps': gbps, 'gpu_name_real': gpu_name_real, 'vkfft_str': vkfft_str,
               'algo_str': algo_str, 'nup_str': nup_str}
    if q is None:
        return results
    else:
        q.put(results)


def bench_pyvkfft_cuda(sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None, args=None, serial=False,
                       inplace=True, r2c=False, dct=False, dst=False):
    if serial:
        return _bench_pyvkfft_cuda(None, sh, precision, ndim, nb_repeat, nb_loop, gpu_name, args, inplace=inplace,
                                   r2c=r2c, dct=dct, dst=dst)
    q = Queue()
    p = Process(target=_bench_pyvkfft_cuda, args=(q, sh, precision, ndim, nb_repeat, nb_loop, gpu_name,
                                                  args, inplace, r2c, dct, dst))
    p.start()
    try:
        results = q.get(timeout=20)
    except:
        results = {'dt': 0, 'gbps': 0, 'gpu_name_real': None, 'vkfft_str': None}
    p.join()
    return results


def _bench_cupy(q, sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None):
    import cupy as cp
    import cupyx
    dtype = np.complex128 if precision == 'double' else np.complex64
    if gpu_name is None:
        dev = cp.cuda.Device(0).use()
    else:
        for i in range(cp.cuda.runtime.getDeviceCount()):
            if gpu_name.lower() in cp.cuda.runtime.getDeviceProperties(i)['name'].decode().lower():
                dev = cp.cuda.Device(i).use()
                break
    gpu_name_real = cp.cuda.runtime.getDeviceProperties(dev.id)['name'].decode()
    d = cp.random.uniform(0, 1, sh, dtype=np.float32).astype(dtype)
    dt = 0
    start = cp.cuda.Event()
    stop = cp.cuda.Event()
    ax = list(range(len(sh)))[-ndim:]
    # Explicitly creating the plan does not speed up
    # plan = cupyx.scipy.fft.get_fft_plan(d, axes=ax, value_type='C2C')
    for i in range(nb_repeat):
        dev.synchronize()
        start.record()
        for ii in range(nb_loop):
            cupyx.scipy.fft.fftn(d, axes=ax, overwrite_x=True)
            cupyx.scipy.fft.ifftn(d, axes=ax, overwrite_x=True)
        # a = cp.fft.fftn(d, axes=ax)
        # a = cp.fft.ifftn(d, axes=ax)
        stop.record()
        dev.synchronize()
        dt1 = cp.cuda.get_elapsed_time(start, stop) / 1000 / nb_loop
        if dt == 0:
            dt = dt1
        elif dt1 < dt:
            dt = dt1
    gbps = d.nbytes * ndim * 2 * 2 / dt / 1024 ** 3
    results = {'dt': dt, 'gbps': gbps, 'gpu_name_real': gpu_name_real}
    if q is None:
        return results
    else:
        q.put(results)


def bench_cupy(sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None, serial=False):
    if serial:
        return _bench_cupy(None, sh, precision, ndim, nb_repeat, nb_loop, gpu_name)
    q = Queue()
    p = Process(target=_bench_cupy, args=(q, sh, precision, ndim, nb_repeat, nb_loop, gpu_name))
    p.start()
    try:
        results = q.get(timeout=10)
    except:
        results = {'dt': 0, 'gbps': 0, 'gpu_name_real': None}
    p.join()
    return results


def _bench_skcuda(q, sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None):
    import pycuda.autoprimaryctx  # See https://github.com/lebedov/scikit-cuda/issues/330#issuecomment-1125471345
    import pycuda.driver as cu_drv
    from pycuda import curandom
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        import skcuda.fft as cu_fft
    dtype = np.complex128 if precision == 'double' else np.complex64
    if gpu_name is None:
        gpu_name_real = pycuda.autoprimaryctx.device.name()
        cu_ctx = pycuda.autoprimaryctx.context
    else:
        for i in range(cu_drv.Device.count()):
            d = cu_drv.Device(i)
            if gpu_name.lower() in d.name().lower():
                gpu_name_real = d.name()
                cu_ctx = d.retain_primary_context()
                break
    cu_ctx.push()
    d = curandom.rand(shape=sh, dtype=np.float32).astype(dtype)
    if ndim == 1:
        plan = cu_fft.Plan(sh[-1], dtype, dtype, batch=sh[-2])
    elif ndim == 2:
        plan = cu_fft.Plan(sh[-2:], dtype, dtype, batch=sh[-3])
    else:
        plan = cu_fft.Plan(sh[-3:], dtype, dtype, batch=sh[-4])
    dt = 0
    start = cu_drv.Event()
    stop = cu_drv.Event()
    for i in range(nb_repeat):
        cu_ctx.synchronize()
        start.record()
        for ii in range(nb_loop):
            cu_fft.fft(d, d, plan)
            cu_fft.ifft(d, d, plan)
        stop.record()
        cu_ctx.synchronize()
        dt1 = stop.time_since(start) / 1000 / nb_loop
        if dt == 0:
            dt = dt1
        elif dt1 < dt:
            dt = dt1
    gbps = d.nbytes * ndim * 2 * 2 / dt / 1024 ** 3
    cu_ctx.pop()
    results = {'dt': dt, 'gbps': gbps, 'gpu_name_real': gpu_name_real}
    if q is None:
        return results
    else:
        q.put(results)


def bench_skcuda(sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None, serial=False):
    if serial:
        return _bench_skcuda(None, sh, precision, ndim, nb_repeat, nb_loop, gpu_name)
    q = Queue()
    p = Process(target=_bench_skcuda, args=(q, sh, precision, ndim, nb_repeat, nb_loop, gpu_name))
    p.start()
    try:
        results = q.get(timeout=10)
    except:
        results = {'dt': 0, 'gbps': 0, 'gpu_name_real': None}
    p.join()
    return results


def _bench_gpyfft(q, sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None, opencl_platform=None):
    results = {'dt': 0, 'gbps': 0, 'gpu_name_real': None, 'platform_name_real': None}
    if max(primes(sh[-1])) > 13:
        q.put(results)
    else:
        import pyopencl as cl
        from pyopencl import clrandom
        import gpyfft
        dtype = np.complex128 if precision == 'double' else np.complex64
        gpu_name_real = gpu_name
        platform_name_real = opencl_platform
        if 'PYOPENCL_CTX' in os.environ:
            cl_ctx = cl.create_some_context()
        else:
            cl_ctx = None
            for p in cl.get_platforms():
                if opencl_platform is not None:
                    if opencl_platform.lower() not in p.name.lower():
                        continue
                for d in p.get_devices():
                    if gpu_name is not None:
                        if gpu_name.lower() not in d.name.lower():
                            continue
                    if d.type & cl.device_type.GPU == 0:
                        continue
                    gpu_name_real = d.name
                    platform_name_real = p.name
                    # print("Selected OpenCL device: ", d.name)
                    cl_ctx = cl.Context(devices=(d,))
                    break
                if cl_ctx is not None:
                    break
        cq = cl.CommandQueue(cl_ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)
        dt = 0
        d = clrandom.rand(cq, shape=sh, dtype=np.float32).astype(dtype)
        for axes in permutations([-1, -2, -3][:ndim]):
            # Shuffle axes order to find fastest transform
            gpyfft_plan = gpyfft.FFT(cl_ctx, cq, d, None, axes=axes)
            for i in range(nb_repeat):
                cq.finish()
                t0 = timeit.default_timer()
                for ii in range(nb_loop):
                    gpyfft_plan.enqueue(forward=True)
                    gpyfft_plan.enqueue(forward=False)
                cq.finish()
                dt1 = (timeit.default_timer() - t0) / nb_loop
                if dt == 0:
                    dt = dt1
                elif dt1 < dt:
                    dt = dt1
            del gpyfft_plan
        gbps = d.nbytes * ndim * 2 * 2 / dt / 1024 ** 3
        results = {'dt': dt, 'gbps': gbps, 'gpu_name_real': gpu_name_real,
                   'platform_name_real': platform_name_real}
        if q is None:
            return results
        else:
            q.put(results)


def bench_gpyfft(sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, gpu_name=None, opencl_platform=None,
                 serial=False):
    if serial:
        return _bench_gpyfft(None, sh, precision, ndim, nb_repeat, nb_loop, gpu_name, opencl_platform)
    q = Queue()
    p = Process(target=_bench_gpyfft, args=(q, sh, precision, ndim, nb_repeat, nb_loop, gpu_name, opencl_platform))
    p.start()
    try:
        results = q.get(timeout=10)
    except:
        results = {'dt': 0, 'gbps': 0, 'gpu_name_real': None, 'platform_name_real': None}
    p.join()
    return results


def _bench_scipy(q, sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, nthreads=None):
    from scipy.fft import fftn, ifftn
    dtype = np.complex128 if precision == 'double' else np.complex64
    gpu_name_real = get_cpu_name()
    # Adjust the batch size for CPU calculations (10-50x slower)
    sh = list(sh)
    sh[0] = max(1, sh[0] // 10)
    d = np.random.uniform(0, 1, sh).astype(dtype)
    dt = 0
    ax = list(range(len(sh)))[-ndim:]
    for i in range(nb_repeat):
        t0 = timeit.default_timer()
        for ii in range(nb_loop):
            a = fftn(d, axes=ax, overwrite_x=True, workers=nthreads)
            a = ifftn(d, axes=ax, overwrite_x=True, workers=nthreads)
        dt1 = (timeit.default_timer() - t0) / nb_loop
        if dt == 0:
            dt = dt1
        elif dt1 < dt:
            dt = dt1
    gbps = d.nbytes * ndim * 2 * 2 / dt / 1024 ** 3
    results = {'dt': dt, 'gbps': gbps, 'gpu_name_real': gpu_name_real}
    if q is None:
        return results
    else:
        q.put(results)


def bench_scipy(sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, nthreads=None, serial=False):
    if serial:
        return _bench_scipy(None, sh, precision, ndim, nb_repeat, nb_loop, nthreads)
    q = Queue()
    p = Process(target=_bench_scipy, args=(q, sh, precision, ndim, nb_repeat, nb_loop, nthreads))
    p.start()
    try:
        results = q.get(timeout=10)
    except:
        results = {'dt': 0, 'gbps': 0, 'gpu_name_real': 'CPU'}
    p.join()
    return results


def _bench_pyfftw(q, sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, nthreads=None):
    from pyfftw.builders import fftn, ifftn

    dtype = np.complex128 if precision == 'double' else np.complex64
    gpu_name_real = get_cpu_name()
    # Adjust the batch size for CPU calculations (10-50x slower)
    sh = list(sh)
    sh[0] = max(1, sh[0] // 10)
    d = np.random.uniform(0, 1, sh).astype(dtype)
    dt = 0
    ax = list(range(len(sh)))[-ndim:]
    t1 = fftn(d, axes=ax, overwrite_input=True, planner_effort='FFTW_MEASURE', threads=nthreads)
    t2 = ifftn(d, axes=ax, overwrite_input=True, planner_effort='FFTW_MEASURE', threads=nthreads)
    for i in range(nb_repeat):
        t0 = timeit.default_timer()
        for ii in range(nb_loop):
            a = t1(d)
            a = t2(d)
        dt1 = (timeit.default_timer() - t0) / nb_loop
        if dt == 0:
            dt = dt1
        elif dt1 < dt:
            dt = dt1
    gbps = d.nbytes * ndim * 2 * 2 / dt / 1024 ** 3
    results = {'dt': dt, 'gbps': gbps, 'gpu_name_real': gpu_name_real}
    if q is None:
        return results
    else:
        q.put(results)


def bench_pyfftw(sh, precision='single', ndim=1, nb_repeat=3, nb_loop=1, nthreads=None, serial=False):
    if serial:
        return _bench_pyfftw(None, sh, precision, ndim, nb_repeat, nb_loop, nthreads)
    q = Queue()
    p = Process(target=_bench_pyfftw, args=(q, sh, precision, ndim, nb_repeat, nb_loop, nthreads))
    p.start()
    try:
        results = q.get(timeout=10)
    except:
        results = {'dt': 0, 'gbps': 0, 'gpu_name_real': 'CPU'}
    p.join()
    return results


def plot_benchmark(results, ndim, gpu_name_real, radix_max, legend_loc="lower left",
                   fig=None, figsize=(16, 8)):
    if fig is None:
        plt.figure(figsize=figsize)
    else:
        plt.clf()
    if radix_max > 13:
        # Use a different symbol for Bluestein
        x = results['n']
        maxprime = np.array(results['maxprime'])
        idx7a = np.where(maxprime <= 7)[0]
        idx7b = np.where(maxprime > 7)[0]
        idx13a = np.where(maxprime <= 13)[0]
        idx13b = np.where(maxprime > 13)[0]
        if "gpyfft[clFFT]" in results:
            y = results["gpyfft[clFFT]"]
            plt.plot(np.take(x, idx13a), np.take(y, idx13a), color='#00FF0D', marker='o', markersize=3, linestyle='',
                     label="gpyfft[clFFT]")
        if "skcuda[cuFFT]" in results:
            y = results["skcuda[cuFFT]"]
            plt.plot(np.take(x, idx7a), np.take(y, idx7a), color='#0073FF', marker='o', markersize=3,
                     linestyle='', label="skcuda[cuFFT] (radix-7)")
            plt.plot(np.take(x, idx7b), np.take(y, idx7b), color='#0073FF', marker='+', markersize=3,
                     linestyle='', label="skcuda[cuFFT] (Bluestein/?)")
        if "vkFFT.opencl" in results:
            y = results["vkFFT.opencl"]
            plt.plot(np.take(x, idx13a), np.take(y, idx13a), color='#FF00F2', marker='o', markersize=3,
                     linestyle='', label="vkFFT.opencl (radix-13)")
            plt.plot(np.take(x, idx13b), np.take(y, idx13b), color='#FF00F2', marker='+', markersize=3,
                     linestyle='', label="vkFFT.opencl (Bluestein)")
        if "vkFFT.cuda" in results:
            y = results["vkFFT.cuda"]
            plt.plot(np.take(x, idx13a), np.take(y, idx13a), color='#FF8C00', marker='o', markersize=3,
                     linestyle='', label="vkFFT.cuda (radix-13)")
            plt.plot(np.take(x, idx13b), np.take(y, idx13b), color='#FF8C00', marker='+', markersize=3,
                     linestyle='', label="vkFFT.cuda (Bluestein)")
    else:
        x = results['n']
        if "gpyfft[clFFT]" in results:
            y = results["gpyfft[clFFT]"]
            plt.plot(x, y, color='#00FF0D', marker='o', markersize=3, linestyle='', label="gpyfft[clFFT]")
        if "skcuda[cuFFT]" in results:
            y = results["skcuda[cuFFT]"]
            plt.plot(x, y, color='#0073FF', marker='o', markersize=3, linestyle='', label="skcuda[cuFFT]")
        if "vkFFT.opencl" in results:
            y = results["vkFFT.opencl"]
            plt.plot(x, y, color='#FF00F2', marker='o', markersize=3, linestyle='', label="vkFFT.opencl")
        if "vkFFT.cuda" in results:
            y = results["vkFFT.cuda"]
            plt.plot(x, y, color='#FF8C00', marker='o', markersize=3, linestyle='', label="vkFFT.cuda")

    plt.legend(loc=legend_loc, fontsize=10)
    plt.xlabel("FFT size", fontsize=12)
    plt.ylabel("idealised throughput [Gbytes/s]", fontsize=12)
    if "skcuda[cuFFT]" in results:
        has_skcuda, cufft_version = test_skcuda()
        cufft_version = ", cuFFT " + cufft_version
    else:
        cufft_version = ""
    if "vkFFT.cuda" in results:
        has_pyvkfft_cuda, cu_version_compile, cu_version_driver, cu_version_runtime = test_pyvkfft_cuda()
        cu_version = ", CUDA driver %s runtime %s" % (cu_version_driver, cu_version_runtime)
    else:
        cu_version = ""
    vkfft_git_v = '' if 'unknown' in vkfft_git_version() else f'[{vkfft_git_version()}]'
    plt.suptitle("%dD FFT speed [%s, VkFFT %s%s%s%s]" % (ndim, gpu_name_real, vkfft_version(),
                                                         vkfft_git_v, cu_version, cufft_version),
                 fontsize=12)
    plt.title("Batched FFTs, 'Ideal' throughput assumes one r+w operation per FFT axis [%s, %s]" %
              (platform.platform(), platform.node()), fontsize=10)
    plt.grid(which='both', alpha=0.3)
    plt.xlim(0)
    plt.ylim(0)
    plt.tight_layout()

    # Force refresh
    plt.draw()
    plt.gcf().canvas.draw()
    plt.pause(.001)


def init_results(has_pyvkfft_opencl, has_pyvkfft_cuda, has_skcuda, has_gpyfft):
    results = {"n": [], "maxprime": []}
    if "vkFFT.opencl" not in results and has_pyvkfft_opencl:
        results["vkFFT.opencl"] = []
        results["vkFFT.opencl-dt"] = []
    if "gpyfft[clFFT]" not in results and has_gpyfft:
        results["gpyfft[clFFT]"] = []
        results["gpyfft[clFFT]-dt"] = []
    if "vkFFT.cuda" not in results and has_pyvkfft_cuda:
        results["vkFFT.cuda"] = []
        results["vkFFT.cuda-dt"] = []
    if "skcuda[cuFFT]" not in results and has_skcuda:
        results["skcuda[cuFFT]"] = []
        results["skcuda[cuFFT]-dt"] = []
    return results


def run(nmin, nmax, radix_max, ndim, precision="single", nb_repeat=3, nb_loop=1, gpu_name=None,
        batch=True, opencl_platform=None, figsize=(16, 8),
        has_pyvkfft_opencl=None, has_pyvkfft_cuda=None, has_gpyfft=None, has_skcuda=None,
        r2c=False, dct=False, dst=False, inplace=True):
    """
    Run the benchmark, measuring the idealised memory throughput (assuming a single
    read+write operation per axis) for an inplace C2C transform using different
    fft backends available.
    Note that each test is made in a separate individual process, so this can
    take a long time.

    :param nmin: smallest size N of the array, e.g. with a shape (batch, N, N)
        for a 2D transform.
    :param nmax: largest size N for the array.
    :param radix_max: maximum radix for the tested sizes. Use a large value (1e7)
        to test all sizes regardless of the prime decomposition.
    :param precision: either 'single' or 'double'
    :param nb_repeat: number of times each fft+ifft cycle is performed, the best
        timing is kept
    :param gpu_name: name or substring (case-insensitive) of the GPU to use.
        If None, the first found will be used.
    :param batch: if True (the default), all transforms are batched so that
        the array size is large enough to yield a measurable transform time. Each
        array takes a shape e.g. (batch, N, N) for a 2D transform.
    :param opencl_platform: name or substring (case-insensitive) of the OpenCL
        platform to use. If None, the first found will be used.
    :param figsize: figure size for plotting. Set to None to disable plotting.
    :param has_pyvkfft_opencl: if True, will test pvkfft.opencl. If None,
        will be automatically detected
    :param has_pyvkfft_cuda: if True, will test pvkfft.cuda. If None,
        will be automatically detected
    :param has_gpyfft: if True, will test gpyfft (clFFT). If None,
        will be automatically detected
    :param has_skcuda: if True, will test scikit.cuda (cuFFT). If None,
        will be automatically detected
    :param r2c: if True, test an r2c transform
    :param dct: test DCT of type 1,2,3 or 4
    :param dst: test DST of type 1,2,3 or 4
    :param inplace: if True, test inplace transforms
    """
    if has_pyvkfft_opencl is None:
        has_pyvkfft_opencl = test_pyvkfft_opencl()
    if has_pyvkfft_cuda is None:
        has_pyvkfft_cuda, cu_version_compile, cu_version_driver, cu_version_runtime = test_pyvkfft_cuda()
    if has_skcuda is None:
        has_skcuda, cufft_version = test_skcuda()
    if has_gpyfft is None:
        has_gpyfft = test_gpyfft()

    results = init_results(has_pyvkfft_opencl, has_pyvkfft_cuda, has_skcuda, has_gpyfft)
    if ndim == 1:
        header_results = "    1 x batch x     N [%dD]" % (ndim)
    elif ndim == 2:
        header_results = "batch x     N x     N [%dD]" % (ndim)
    else:
        header_results = "batch x    N x    N x    N [%dD]" % (ndim)
    for b in results.keys():
        if b not in ["n", "maxprime"] and "-dt" not in b:
            header_results += "%17s  " % b

    s = f"DCT{dct}" if dct else f"DST{dst}" if dst else "R2C" if r2c else "C2C"
    if inplace:
        s = "inplace " + s

    print(f"Gbytes/s and time given for a couple (FFT, iFFT) of {s}, "
          f"dtype={np.dtype(np.complex64).name}")
    print()
    print(header_results)

    if figsize is not None:
        fig = plt.figure(figsize=figsize)

    dtype = np.complex128 if precision == 'double' else np.complex64
    gpu_name_real_ok = None
    gpu_name_real = None

    for n in range(nmin, nmax + 1):
        maxprime = max(primes(n))
        if maxprime > radix_max:
            continue
        results["n"].append(n)
        results["maxprime"].append(maxprime)
        if batch:
            # Estimate batch size to last 0.05s with at least 100 GB/s
            nb = int(round(0.05 * 100 / (n ** ndim * np.dtype(dtype).itemsize * ndim * 2 * 2 / 1024 ** 3)))
            nb = max(nb, 1)
            nb = min(nb, 99999)
        else:
            nb = 1

        if ndim == 1:
            sh = 1, nb, n
        elif ndim == 2:
            sh = nb, n, n
        else:
            sh = nb, n, n, n

        vkfft_str = None
        # OpenCL backends
        if has_pyvkfft_opencl:
            res = bench_pyvkfft_opencl(sh, precision, ndim, nb_repeat, nb_loop, gpu_name, opencl_platform,
                                       r2c=r2c, dct=dct, dst=dst, inplace=inplace)
            dt = res['dt']
            gbps = res['gbps']
            gpu_name_real = res['gpu_name_real']
            platform_name_real = res['platform_name_real']
            vkfft_str = res['vkfft_str']
            results["vkFFT.opencl"].append(gbps)
            results["vkFFT.opencl-dt"].append(dt)

        if gpu_name_real_ok is None and gpu_name_real is not None:
            gpu_name_real_ok = gpu_name_real

        if has_gpyfft:
            res = bench_gpyfft(sh, precision, ndim, nb_repeat, nb_loop, gpu_name, opencl_platform)
            dt = res['dt']
            gbps = res['gbps']
            gpu_name_real = res['gpu_name_real']
            platform_name_real = res['platform_name_real']
            results["gpyfft[clFFT]"].append(gbps)
            results["gpyfft[clFFT]-dt"].append(dt)

        # CUDA backends
        if has_pyvkfft_cuda:
            res = bench_pyvkfft_cuda(sh, precision, ndim, nb_repeat, nb_loop, gpu_name)
            dt = res['dt']
            gbps = res['gbps']
            gpu_name_real = res['gpu_name_real']
            vkfft_str = res['vkfft_str']
            results["vkFFT.cuda"].append(gbps)
            results["vkFFT.cuda-dt"].append(dt)

        if gpu_name_real_ok is None and gpu_name_real is not None:
            gpu_name_real_ok = gpu_name_real

        if has_skcuda:
            res = bench_skcuda(sh, precision, ndim, nb_repeat, nb_loop, gpu_name)
            dt = res['dt']
            gbps = res['gbps']
            gpu_name_real = res['gpu_name_real']
            results["skcuda[cuFFT]"].append(gbps)
            results["skcuda[cuFFT]-dt"].append(dt)

        # text output
        if ndim == 3:
            r = " %4d x %4d x %4d x %4d      " % sh
        else:
            r = "%5d x %5d x %5d      " % sh
        for b in results.keys():
            if b not in ["n", "maxprime"] and "-dt" not in b:
                dt = results[b + '-dt'][-1] / nb
                if dt < 1e-3:
                    r += "%7.2f [%6.2f µs]" % (results[b][-1], dt * 1e6)
                elif dt > 1:
                    r += "%7.2f [%6.2f  s]" % (results[b][-1], dt)
                else:
                    r += "%7.2f [%6.2f ms]" % (results[b][-1], dt * 1000)
        if vkfft_str is not None:
            r += f" [VkFFT: {vkfft_str}]"
        print(r)

        if len(results['n']) % 10 == 9 and figsize is not None:
            plot_benchmark(results, ndim, gpu_name_real_ok, radix_max, legend_loc="upper right",
                           fig=fig)

    if figsize is not None:
        plot_benchmark(results, ndim, gpu_name_real_ok, radix_max, "upper right", fig=fig)
        r = "-radix%d" % radix_max if radix_max < 100 else ""
        figname = 'benchmark-%dDFFT%s-%s-%s-%s-%s.png' % (ndim, gpu_name_real_ok.replace(' ', '_'), r,
                                                          platform.platform(), platform.node(),
                                                          strftime("%Y-%m-%d-%Hh%M", localtime()))
        plt.savefig(figname)
        print("Saved benchmark figure to: \n    %s" % figname)
    print()
    return results


if __name__ == '__main__':
    # res = run(nmin=32, nmax=256, radix_max=7, ndim=2, gpu_name=None)
    # res = run(nmin=32, nmax=256, radix_max=7, ndim=2, gpu_name=None, batch=False)
    # res = run(nmin=128, nmax=256, radix_max=3, ndim=2, gpu_name=None, figsize=None)
    # res = run(nmin=128, nmax=256, radix_max=3, ndim=2, gpu_name=None, figsize=None, r2c=True)
    # res = run(nmin=128, nmax=256, radix_max=3, ndim=2, gpu_name=None, figsize=None, r2c=True, inplace=False)
    res = run(nmin=128, nmax=256, radix_max=3, ndim=2, gpu_name=None, figsize=None, has_gpyfft=False,
              has_skcuda=False)
    res = run(nmin=128, nmax=256, radix_max=3, ndim=2, gpu_name=None, figsize=None, has_gpyfft=False,
              has_skcuda=False, r2c=True)
    res = run(nmin=128, nmax=256, radix_max=3, ndim=2, gpu_name=None, figsize=None, has_gpyfft=False,
              has_skcuda=False, r2c=True, inplace=False)



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2021- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr

import warnings
import ctypes
import numpy as np
import pyopencl as cl
import pyopencl.array as cla
from .base import load_library, VkFFTApp as VkFFTAppBase, check_vkfft_result, ctype_int_size_p
from . import config
from .tune import tune_vkfft

try:
    _vkfft_opencl = load_library("_vkfft_opencl")


    class _types:
        """Aliases"""
        vkfft_config = ctypes.c_void_p
        vkfft_app = ctypes.c_void_p


    _vkfft_opencl.make_config.restype = ctypes.c_void_p
    _vkfft_opencl.make_config.argtypes = [ctype_int_size_p, ctypes.c_size_t,
                                          ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p,
                                          ctypes.c_void_p, ctypes.c_int, ctypes.c_size_t, ctypes.c_int,
                                          ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                          ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                          ctypes.c_size_t, ctype_int_size_p, ctypes.c_int,
                                          ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                          ctypes.c_int, ctype_int_size_p, ctypes.c_int]

    _vkfft_opencl.init_app.restype = ctypes.c_void_p
    _vkfft_opencl.init_app.argtypes = [_types.vkfft_config, ctypes.c_void_p,
                                       ctypes.POINTER(ctypes.c_int),
                                       ctypes.POINTER(ctypes.c_size_t),
                                       ctype_int_size_p, ctype_int_size_p]

    _vkfft_opencl.fft.restype = ctypes.c_int
    _vkfft_opencl.fft.argtypes = [_types.vkfft_app, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]

    _vkfft_opencl.ifft.restype = ctypes.c_int
    _vkfft_opencl.ifft.argtypes = [_types.vkfft_app, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]

    _vkfft_opencl.free_app.restype = None
    _vkfft_opencl.free_app.argtypes = [_types.vkfft_app]

    _vkfft_opencl.free_config.restype = None
    _vkfft_opencl.free_config.argtypes = [_types.vkfft_config]

    _vkfft_opencl.vkfft_version.restype = ctypes.c_uint32
    _vkfft_opencl.vkfft_version.argtypes = None

    _vkfft_opencl.vkfft_max_fft_dimensions.restype = ctypes.c_uint32
    _vkfft_opencl.vkfft_max_fft_dimensions.argtypes = None
except OSError:
    # This is used for doc generation
    import sys

    if 'sphinx' in sys.modules:
        pass
    else:
        raise


class VkFFTApp(VkFFTAppBase):
    """
    VkFFT application interface implementing a FFT plan.
    """

    def __init__(self, shape, dtype: type, queue: cl.CommandQueue, ndim=None, inplace=True, norm=1,
                 r2c=False, dct=False, dst=False, axes=None, strides=None, tune_config=None,
                 r2c_odd=False, verbose=False, **kwargs):
        """
        Init function for the VkFFT application.

        :param shape: the shape of the array to be transformed. The number
            of dimensions of the array can be larger than the FFT dimensions.
        :param dtype: the numpy dtype of the source array (can be complex64 or complex128)
        :param queue: the pyopencl CommandQueue to use for the transform.
        :param ndim: the number of dimensions to use for the FFT. By default,
            uses the array dimensions. Can be smaller, e.g. ndim=2 for a 3D
            array to perform a batched 3D FFT on all the layers. The FFT
            is always performed along the last axes if the array's number
            of dimension is larger than ndim, i.e. on the x-axis for ndim=1,
            on the x and y axes for ndim=2, etc.. Unless axes are given.
        :param inplace: if True (the default), performs an inplace transform and
            the destination array should not be given in fft() and ifft().
        :param norm: if 0 (unnormalised), every transform multiplies the L2
            norm of the array by its size (or the size of the transformed
            array if ndim<d.ndim).
            if 1 (the default) or "backward", the inverse transform divides
            the L2 norm by the array size, so FFT+iFFT will keep the array norm.
            if "ortho", each transform will keep the L2 norm, but that will
            involve an extra read & write operation.
        :param r2c: if True, will perform a real->complex transform, where the
            complex destination is a half-hermitian array.
            For an inplace transform, if the transformed data shape is (...,nx),
            the input float array should have a shape of (..., nx+2), the last
            two columns being ignored in the input data, and the resulting
            complex array (using pycuda's GPUArray.view(dtype=np.complex64) to
            reinterpret the type) will have a shape (..., nx//2 + 1).
            For an out-of-place transform, if the input (real) shape is (..., nx),
            the output (complex) shape should be (..., nx//2+1).
            Note that for C2R transforms with ndim>=2, the source (complex) array
            is modified.
            For an inplace transform with an odd-sized x-axis, see the r2c_odd
            parameter.
        :param dct: used to perform a Direct Cosine Transform (DCT) aka a R2R transform.
            An integer can be given to specify the type of DCT (1, 2, 3 or 4).
            if dct=True, the DCT type 2 will be performed, following scipy's convention.
        :param dst: used to perform a Direct Sine Transform (DST) aka a R2R transform.
            An integer can be given to specify the type of DST (1, 2, 3 or 4).
            if dst=True, the DST type 2 will be performed, following scipy's convention.
        :param axes: a list or tuple of axes along which the transform should be made.
            if None, the transform is done along the ndim fastest axes, or all
            axes if ndim is None. For R2C transforms, the fast axis must be
            transformed.
        :param strides: the array strides - needed if not C-ordered.
        :param tune_config: this can be used to automatically generate an
            optimised set of VkFFT parameters by testing various configurations
            and measuring the FFT speed, in a manner similar to fftw's FFTW_MEASURE.
            This should be a dictionary including the backend used and the parameter
            values which will be tested.
            This is EXPERIMENTAL, as wrong parameters may lead to crashes.
            Note that this will allocate temporary GPU arrays, unless the arrays
            to used have been passed as parameters ('dest' and 'src').
            Examples:
            tune={'backend':'cupy} - minimal example, will automatically test a small
            set of parameters (4 to 9 tests), depending on the GPU type.
            tune={'backend':'cupy, 'warpSize':[8,16,32,64,128]}: this will test
            5 possible values for the warpSize.
            tune={'backend':'cupy, 'groupedBatch':[[-1,-1,-1],[8,8,8], [4,16,16}:
            this will test 3 possible values for groupedBatch. This one is more
            tricky to use.
            tune={'backend':'cupy, 'warpSize':[8,16,32,64,128], 'src':a}: this
            will test 5 possible values for the warpSize, with a given source GPU
            array. This would only be valid for an inplace transform as no
            destination array is given.
        :param r2c_odd: this should be set to True to perform an inplace r2c/c2r
            transform with an odd-sized fast (x) axis.
            Explanation: to perform a 1D inplace transform of an array with 100
            elements, the input array should have a 100+2 size, resulting in
            a half-Hermitian array of size 51. If the input data has a size
            of 101, the input array should also be padded to 102 (101+1), and
            the resulting half-Hermitian array also has a size of 51. A
            flag is thus needed to differentiate the cases of 100+2 or 101+1.
        :param verbose: if True, print a 1-string info about this VkFFTApp.
            See __str__ for details.

        :raises RuntimeError: if the initialisation fails, e.g. if the GPU
            driver has not been properly initialised, or if the transform dimensions
            or data type are not allowed by VkFFT.
        """
        if tune_config is not None:
            kwargs = tune_vkfft(tune_config, shape=shape, dtype=dtype, ndim=ndim, queue=queue, inplace=inplace,
                                norm=norm, r2c=r2c, dct=dct, dst=dst, axes=axes, strides=strides, verbose=False,
                                r2c_odd=r2c_odd, **kwargs)[0]
        super().__init__(shape, dtype, ndim=ndim, inplace=inplace, norm=norm, r2c=r2c,
                         dct=dct, dst=dst, axes=axes, strides=strides, r2c_odd=r2c_odd, **kwargs)
        self.queue = queue

        if self.precision == 2 and 'cl_khr_fp16' not in self.queue.device.extensions:
            raise RuntimeError("Half precision required but cl_khr_fp16 extension is not available")
        if self.precision == 8 and 'cl_khr_fp64' not in self.queue.device.extensions:
            raise RuntimeError("Double precision required but cl_khr_fp64 extension is not available")

        self.config = self._make_config()

        if self.config is None:
            raise RuntimeError("Error creating VkFFTConfiguration. Was the OpenCL context properly initialised ?")

        res = ctypes.c_int(0)
        # Size of tmp buffer allocated by VkFFT - if any
        tmp_buffer_nbytes = ctypes.c_size_t(0)
        # 0 or 1 for each axis, only if the Bluestein algorithm is used
        use_bluestein_fft = np.zeros(vkfft_max_fft_dimensions(), dtype='long')
        # number of axis upload per dimension
        num_axis_upload = np.zeros(vkfft_max_fft_dimensions(), dtype='long')

        self.app = _vkfft_opencl.init_app(self.config, queue.int_ptr, ctypes.byref(res),
                                          ctypes.byref(tmp_buffer_nbytes),
                                          use_bluestein_fft, num_axis_upload)
        try:
            check_vkfft_result(res, shape=shape, dtype=dtype, ndim=ndim, inplace=inplace, norm=norm,
                               r2c=r2c, dct=dct, dst=dst, axes=axes,
                               backend="opencl:%s:%s" % (queue.device.platform.name, queue.device.name),
                               strides=self.strides0,
                               vkfft_shape=self.shape, vkfft_skip=self.skip_axis, vkfft_nbatch=self.n_batch)
        except Exception as ex:
            print("clVkFFTApp error:", shape, strides, dtype, ndim, axes, self.shape, self.skip_axis)
            raise ex
        if self.app is None:
            raise RuntimeError("Error creating VkFFTApplication. Was the OpenCL context properly initialised ?")

        self.tmp_buffer_nbytes = np.int64(tmp_buffer_nbytes)
        self.use_bluestein_fft = [bool(n) for n in use_bluestein_fft[:len(self.shape)]]
        self.nb_axis_upload = [int(num_axis_upload[i] * (self.skip_axis[i] is False))
                               for i in range(len(self.shape))]

        if verbose:
            print(self)

    def __del__(self):
        """ Takes care of deleting allocated memory in the underlying
        VkFFTApplication and VkFFTConfiguration.
        """
        if self.app is not None:
            _vkfft_opencl.free_app(self.app)
        if self.config is not None:
            _vkfft_opencl.free_config(self.config)

    def _make_config(self):
        """ Create a vkfft configuration for a FFT transform"""
        if len(self.shape) > vkfft_max_fft_dimensions():
            raise RuntimeError(f"Too many FFT dimensions after collapsing non-transform axes: "
                               f"{len(self.shape)}>{vkfft_max_fft_dimensions()}")

        shape = np.ones(vkfft_max_fft_dimensions(), dtype='long')
        shape[:len(self.shape)] = self.shape

        skip = np.zeros(vkfft_max_fft_dimensions(), dtype='long')
        skip[:len(self.skip_axis)] = self.skip_axis

        grouped_batch = np.empty(vkfft_max_fft_dimensions(), dtype='long')
        grouped_batch.fill(-1)
        grouped_batch[:len(self.groupedBatch)] = self.groupedBatch

        if self.r2c and self.inplace:
            # the last one or two columns are ignored in the R array, and will be used
            # in the C array with a size nx//2+1
            if self.r2c_odd:
                shape[0] -= 1
            else:
                shape[0] -= 2

        if self.norm == "ortho":
            norm = 0
        else:
            norm = self.norm

        # We pass fake buffer pointer addresses to VkFFT. The real ones will be
        # given when performing the actual FFT.
        ctx = self.queue.context
        device = ctx.devices[0]
        platform = device.platform
        dest_gpudata = 2
        if self.inplace:
            dest_gpudata = 0

        return _vkfft_opencl.make_config(shape, self.ndim, 1, dest_gpudata, platform.int_ptr,
                                         device.int_ptr, ctx.int_ptr,
                                         norm, self.precision, int(self.r2c), int(self.dct),
                                         int(self.dst),
                                         int(self.disableReorderFourStep), int(self.registerBoost),
                                         int(self.use_lut), int(self.keepShaderCode),
                                         self.n_batch, skip,
                                         int(self.coalescedMemory), int(self.numSharedBanks),
                                         int(self.aimThreads), int(self.performBandwidthBoost),
                                         int(self.registerBoostNonPow2), int(self.registerBoost4Step),
                                         int(self.warpSize), grouped_batch,
                                         int(self.forceCallbackVersionRealTransforms))

    def fft(self, src: cla.Array, dest: cla.Array = None, queue: cl.CommandQueue = None):
        """
        Compute the forward FFT

        :param src: the source pyopencl Array
        :param dest: the destination pyopencl Array. Should be None for an inplace transform
        :param queue: the pyopencl CommandQueue to use for the transform. If not given,
            the queue of the source array is used. If the queue does not match
            the application's, a warning is emitted (see config.WARN_OPENCL_QUEUE_MISMATCH).
            If a queue is not supplied and the source and destination arrays do not have the
            same queue, then a RuntimeError is raised.
        :raises RuntimeError: in case of a GPU kernel launch error
        :return: the transformed array. For a R2C inplace transform, the complex view of the
            array is returned.
        """
        if not queue:
            if dest is None:
                if src.queue is None:
                    warnings.warn("queue is not given and the source array does not have a queue "
                                  "attached to it. Falling back to the queue given to the constructor "
                                  "of VkFFTApp. This is deprecated and will stop working in the future. "
                                  "Use src_array.with_queue(queue) to attach a queue to the array or "
                                  "pass a queue to this method", DeprecationWarning)
                    queue = self.queue
                else:
                    queue = src.queue
            elif dest.queue is None and src.queue is None:
                warnings.warn("queue is not given and the source/dest arrays do not have a queue "
                              "attached to it. Falling back to the queue given to the constructor "
                              "of VkFFTApp. This is deprecated and will stop working in the future. "
                              "Use src_array.with_queue(queue) to attach a queue to the array or "
                              "pass a queue to this method", DeprecationWarning)
                queue = self.queue
            elif dest.queue != src.queue:
                if dest.queue is None:
                    queue = src.queue
                elif src.queue is None:
                    queue = dest.queue
                else:
                    raise RuntimeError("queue is not given and the source/dest arrays queues differ. "
                                       "Please supply a queue when calling fft(...), or "
                                       "use array.with_queue(queue)")
            else:
                queue = src.queue

        if config.WARN_OPENCL_QUEUE_MISMATCH and self.queue != queue:
            warnings.warn("Using the array queue for transform, which differs from the application "
                          "queue. NB: this warning will be removed in a future version , and you can "
                          "suppress it using config.WARN_OPENCL_QUEUE_MISMATCH",
                          UserWarning)

        if self.inplace:
            if dest is not None:
                if src.data.int_ptr != dest.data.int_ptr:
                    raise RuntimeError("VkFFTApp.fft: dest is not None but this is an inplace transform")
            res = _vkfft_opencl.fft(self.app, int(src.data.int_ptr), int(src.data.int_ptr), int(queue.int_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst,
                               backend="opencl:%s:%s" % (queue.device.platform.name, queue.device.name),
                               strides=self.strides0, vkfft_shape=self.shape,
                               vkfft_skip=self.skip_axis, vkfft_nbatch=self.n_batch)
            if self.norm == "ortho":
                src *= self._get_fft_scale(norm=0)
            if self.r2c:
                if src.dtype == np.float32:
                    return src.view(dtype=np.complex64)
                elif src.dtype == np.float64:
                    return src.view(dtype=np.complex128)
            return src
        else:
            if dest is None:
                raise RuntimeError("VkFFTApp.fft: dest is None but this is an out-of-place transform")
            elif src.data.int_ptr == dest.data.int_ptr:
                raise RuntimeError("VkFFTApp.fft: dest and src are identical but this is an out-of-place transform")
            if self.r2c:
                tmp = src.size // src.shape[self.fast_axis] * (src.shape[self.fast_axis] // 2 + 1)
                if dest.size != tmp:
                    raise RuntimeError(f"VkFFTApp.fft the destination array size ({dest.size}) is not "
                                       f"equal to the expected one ({tmp}) for an R2C transform "
                                       f"[fast axis={self.fast_axis}, src shape={src.shape}, "
                                       f"axes={self.axes0}, strides={src.strides}]")
            res = _vkfft_opencl.fft(self.app, int(src.data.int_ptr), int(dest.data.int_ptr), int(queue.int_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst,
                               backend="opencl:%s:%s" % (queue.device.platform.name, queue.device.name),
                               strides=self.strides0, vkfft_shape=self.shape,
                               vkfft_skip=self.skip_axis, vkfft_nbatch=self.n_batch)
            if self.norm == "ortho":
                dest *= self._get_fft_scale(norm=0)
            return dest

    def ifft(self, src: cla.Array, dest: cla.Array = None, queue: cl.CommandQueue = None):
        """
        Compute the backward FFT

        :param src: the source pyopencl.Array
        :param dest: the destination pyopencl.Array. Can be None for an inplace transform
        :param queue: the pyopencl CommandQueue to use for the transform. If not given,
            the queue of the source array is used. If the queue does not match
            the application, a warning is emitted (see config.WARN_OPENCL_QUEUE_MISMATCH).
            If a queue is not supplied and the source and destination arrays do not have the
            same queue, then a RuntimeError is raised.
        :raises RuntimeError: in case of a GPU kernel launch error
        :return: the transformed array. For a C2R inplace transform, the float view of the
            array is returned.
        """
        if not queue:
            if dest is None:
                if src.queue is None:
                    warnings.warn("queue is not given and the source array does not have a queue "
                                  "attached to it. Falling back to the queue given to the constructor "
                                  "of VkFFTApp. This is deprecated and will stop working in the future. "
                                  "Use src_array.with_queue(queue) to attach a queue to the array or "
                                  "pass a queue to this method", DeprecationWarning)
                    queue = self.queue
                else:
                    queue = src.queue
            elif dest.queue is None and src.queue is None:
                warnings.warn("queue is not given and the source/dest arrays do not have a queue "
                              "attached to it. Falling back to the queue given to the constructor "
                              "of VkFFTApp. This is deprecated and will stop working in the future. "
                              "Use src_array.with_queue(queue) to attach a queue to the array or "
                              "pass a queue to this method", DeprecationWarning)
                queue = self.queue
            elif dest.queue != src.queue:
                if dest.queue is None:
                    queue = src.queue
                elif src.queue is None:
                    queue = dest.queue
                else:
                    raise RuntimeError("queue is not given and the source/dest arrays queues differ. "
                                       "Please supply a queue when calling fft(...), or "
                                       "use array.with_queue(queue)")
            else:
                queue = src.queue

        if config.WARN_OPENCL_QUEUE_MISMATCH and self.queue != queue:
            warnings.warn("Using the array queue for transform, which differs from the application "
                          "queue. NB: this warning will be removed in a future version , and you can "
                          "suppress it using config.WARN_OPENCL_QUEUE_MISMATCH",
                          UserWarning)

        if self.inplace:
            if dest is not None:
                if src.data.int_ptr != dest.data.int_ptr:
                    raise RuntimeError("VkFFTApp.fft: dest!=src but this is an inplace transform")
            res = _vkfft_opencl.ifft(self.app, int(src.data.int_ptr), int(src.data.int_ptr), int(queue.int_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst,
                               backend="opencl:%s:%s" % (queue.device.platform.name, queue.device.name),
                               strides=self.strides0, vkfft_shape=self.shape,
                               vkfft_skip=self.skip_axis, vkfft_nbatch=self.n_batch)
            if self.norm == "ortho":
                src *= self._get_ifft_scale(norm=0)
            if self.r2c:
                if src.dtype == np.complex64:
                    return src.view(dtype=np.float32)
                elif src.dtype == np.complex128:
                    return src.view(dtype=np.float64)
            return src
        else:
            if dest is None:
                raise RuntimeError("VkFFTApp.ifft: dest is None but this is an out-of-place transform")
            elif src.data.int_ptr == dest.data.int_ptr:
                raise RuntimeError("VkFFTApp.ifft: dest and src are identical but this is an out-of-place transform")
            if self.r2c:
                assert (src.size == dest.size // dest.shape[self.fast_axis] * (dest.shape[self.fast_axis] // 2 + 1))
                # Special case, src and dest buffer sizes are different,
                # VkFFT is configured to go back to the source buffer
                res = _vkfft_opencl.ifft(self.app, int(dest.data.int_ptr), int(src.data.int_ptr),
                                         int(queue.int_ptr))
            else:
                res = _vkfft_opencl.ifft(self.app, int(src.data.int_ptr), int(dest.data.int_ptr),
                                         int(queue.int_ptr))
            check_vkfft_result(res, src.shape, src.dtype, self.ndim, self.inplace, self.norm, self.r2c,
                               self.dct, self.dst,
                               backend="opencl:%s:%s" % (queue.device.platform.name, queue.device.name),
                               strides=self.strides0, vkfft_shape=self.shape,
                               vkfft_skip=self.skip_axis, vkfft_nbatch=self.n_batch)
            if self.norm == "ortho":
                dest *= self._get_ifft_scale(norm=0)
            return dest


def vkfft_version():
    """
    Get VkFFT version

    :return: version as X.Y.Z
    """
    int_ver = _vkfft_opencl.vkfft_version()
    return "%d.%d.%d" % (int_ver // 10000, (int_ver % 10000) // 100, int_ver % 100)


def vkfft_max_fft_dimensions():
    """
    Get the maximum number of dimensions VkFFT can handle. This is
    set at compile time. VkFFT default is 4, pyvkfft sets this to 8.
    Note that consecutive non-transformed are collapsed into a single
    axis, reducing the effective number of dimensions.

    :return: VKFFT_MAX_FFT_DIMENSIONS
    """
    return _vkfft_opencl.vkfft_max_fft_dimensions()



# -*- coding: utf-8 -*-

# PyVkFFT
#   (c) 2024- : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr
#
#
# Basic test imports of the opencl library
# Primarily used to check build status.

import os
from ..base import _library_path, load_library

assert os.path.exists(_library_path("_vkfft_opencl")), "_vkfft_opencl shared library not found"
# _vkfft_opencl = load_library("_vkfft_opencl")
